# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import sys, os, time
from ..ylsys import py2, tmpYl, sysi
from ..ylcompat import isstr, beforImportPlt, ModuleNotFoundError

from .toolIo import filename


def importAllFunCode(mod=None):
    '''
    mod 为包名(type(mod)=='str')或包本身(type(mod)==module)
    自动生成导入所有模块语句 并过滤掉__name__等
    '''
    if mod is None:
        mod = 'yllab'
    if isinstance(mod,str):
        exec ('import %s as mod'%mod)

    names = [name for name in dir(mod) if not ((len(name)>2 and name[:2]=='__') or 
                                  name in ['unicode_literals',])]
    n = 5
    lines = []
    while len(names) > n:
        l,names = names[:n],names[n:]
        lines += [', '.join(l)]
    lines += [', '.join(names)]
    lines = ',\n          '.join(lines)
    
    strr = (("from %s import *\n    from %s import (%s)"%(mod.__name__,mod.__name__,lines)))
    strr = '''from %s import *
try:
    from %s import (%s)
except ImportError:
    pass'''%(mod.__name__,mod.__name__,lines)
    print(strr)

class impt():
    '''
    Only one import to support both environments : __name__ == '__main__' or in a package.
    And no need relative import.
        
    usage： 
        using: 
            >>> with impt():
            >>>     import local_py
        instead of :
            >>> if __name__ == '__main__':
            >>>     import local_py
            >>> else:
            >>>     from . import local_py
            
    Principle：
        temporary add the relpath in sys.path during with statement
        
    Parameters
    ----------
    relpath: str, default
        the dir path of the .py file that you want to import
        
    Zh-cn: 
        在包内或 __name__ == '__main__' 都能直接导入文件.
    '''
    from multiprocessing import Lock
    lock = Lock() # ensure work fine in multi-threading
    def __init__(self, relpath='.'):
        frame = sys._getframe(1)
        _file_ = frame.f_globals.get('__file__', "python_shell.py")
        dirr = os.path.dirname(_file_)
        self.d = os.path.abspath(os.path.join(dirr, os.path.expanduser(relpath)))
    def __enter__(self):
        with self.lock:
            sys.path.insert(0, self.d)
    def __exit__(self,*l):
        with self.lock:
            if sys.path[0] == self.d:
                assert sys.path.pop(0)==self.d, 'impt sys.path error'
            else:
                ind = sys.path.index(self.d)
                assert sys.path.pop(ind)==self.d, 'impt sys.path error'


class inpkg():
    '''
    inpkg = in package
    
    Execute relative import under __name__ == '__main__' enviroment in a package.
        
    usage： 
        using: 
            >>> with inpkg():
            >>>     from . import local_py
            
    Principle：
        auto search and import "top level package". Then, temporary replace __name__ to "module name under top level package" during with statement
        
    Zh-cn: 
        可以能直接运行包内含有 relative import code 的 py 文件
    '''
    def __init__(self):
        frame = sys._getframe(1)
        self.frame = frame
        self._file_ = frame.f_globals['__file__'] if '__file__' in frame.f_globals else frame.f_code.co_filename
        self._name_ = frame.f_globals['__name__']
        # NOTICE: second time %run will no '__package__' key
        self._package_ = self.frame.f_globals.get('__package__', None)  
        self.importTopLevelPackage = self._name_ == '__main__' or self._name_ == filename(self._file_)
        
    def findPackageRoot(self):
        dirr  = os.path.abspath(self._file_)
        files = []
        while len(dirr) > 1:
            files.append(filename(dirr))
            dirr = os.path.dirname(dirr)
            _init_p = os.path.join(dirr, '__init__.py')
            if not os.path.isfile(_init_p):
                return dirr, files
        raise Exception('Has __init__.py in root "/__init__.py"')
        
    def __enter__(self):
        if self.importTopLevelPackage:
            packageroot, files = self.findPackageRoot()
            if len(files) > 1:
                importByPath(os.path.join(packageroot, files[-1]))
            self.frame.f_globals['__name__'] = '.'.join(files[::-1])
            self.frame.f_globals['__package__'] = '.'.join(files[1:][::-1])
            
    def __exit__(self,*l):
        if self.importTopLevelPackage:
            self.frame.f_globals['__name__'] = self._name_
            if self._package_ is None:
                self.frame.f_globals.pop('__package__')
            else:
                self.frame.f_globals['__package__'] = self._package_


def importByPath(pyPath):
    '''
    import `.py` file by a python file path, return the py file as a moudle

    >>> module = importByPath('far/away.py')
    '''
    from boxx import os, dirname, sys, withfun
    pyFile = pyPath
    assert os.path.isfile(pyFile) or os.path.isdir(pyFile), pyFile
    dirr = dirname(pyFile)
    import importlib
    def exitFun(*l):
        assert sys.path.pop(0)==dirr
    with withfun(lambda :sys.path.insert(0, dirr), exitFun):
        module = importlib.import_module(os.path.basename(pyFile).replace('.py',''))
        return module
    
def tryImport(moduleName):
    '''
    try `import @moduleName`. if @moduleName is not installed, return a FakeModule to placeholder the module name
    '''
    module = None
    try:
        exec('import %s as module' % moduleName)
        return module
    except (ModuleNotFoundError, ImportError):
        return  '''"%s" is not install in your Python Enveroment! 
This is a fake one. Please install "%s" and retry''' % (moduleName, moduleName)
        return FakeModule(moduleName)
__FAKE_DIC__ = {}
class FakeModule():
    '''
    a fake Module to placeholder the module name that some module may not installed.
    once use this module, will raise ImportError
    '''
    def __init__(self, name):
        __FAKE_DIC__[id(self)] = '"%s" is not install in your Python Enveroment! This is a fake one. Please install "%s" and retry' % (name, name)
    def __getattribute__(self, name=None, *l):
        raise ImportError(__FAKE_DIC__[id(self)])
    def __getattr__(self, name):
        raise ImportError(__FAKE_DIC__[id(self)])
    def __setattr__(self, name, v, *l):
        raise ImportError(__FAKE_DIC__[id(self)])
    def __str__(self):
        errorMsg = __FAKE_DIC__[id(self)]
        raise ImportError(errorMsg)
    __repr__ = __str__

def removeImportSelf(modelName='boxx.out'):
    '''
    remove a model while import itself
    that mean every time import model will exec model again
    
    Notice: Only support Python 3, For Python 2 only first time import works 
    '''
    if py2 :
        return 
    sys.modules.pop(modelName)
    
    f = sys._getframe(4)
    if 'spec' in f.f_locals:
        f.f_locals['spec'].name = 'sys'

def removeimp(modulesName='boxx'):
    '''
    remove all module by name
    '''
    [sys.modules.pop(k) for k,v in list(sys.modules.items()) if k.startswith(modulesName + '.') or k == modulesName]

def crun(pycode, snakeviz=True):
    '''
    use snakeviz and cProfile to analyse the code performance
    a visualization flame graph web page will be opened in your web browser
    
    Parameters
    ----------
    pycode : str
        Python code
    snakeviz : bool, default True
        use snakeviz to get flame graph in web page
        otherwise, print cProfile result sorted by time
    '''
    from cProfile import run

    if not snakeviz:
        return run(pycode,sort='time')
    import webbrowser
    if not webbrowser.open(''):
        from boxx import warn
        msg = '''**Can't detect browser** in operating environment.
so, we use cProfile.run(pycode,sort='time'),
instead of using snakeviz to visualization code perfomance in web page'''
        warn(msg)
        run(pycode,sort='time')
        print('\n\n'+msg)
        return 
    run(pycode, os.path.join(tmpYl, "snakeviz.result"))
    if not sysi.win:
        from . import softInPath
        assert softInPath('snakeviz'),'run `pip install snakeviz`'
        os.system('snakeviz %s &'% os.path.join(tmpYl,'snakeviz.result'))
    elif sysi.win:
        os.system('start /b  snakeviz.exe %s '% os.path.join(tmpYl,'snakeviz.result'))
        
    
def performance(pyfileOrCode, snakeviz=True):
    '''
    use snakeviz and cProfile to analyse the python file or python code performance
    a visualization flame graph web page will be opened in your web browser
    
    Parameters
    ----------
    pyfileOrCode : str
        Python file's path or python code
    snakeviz : bool, default True
        use snakeviz to get flame graph in web page
        otherwise, print cProfile result sorted by time
    '''
    if pyfileOrCode.endswith('.py'):
        crun("from boxx import runpyfile;runpyfile('%s')"%pyfileOrCode)
    else:
        crun(pyfileOrCode)

class timeit():
    '''
    usage 1 :
        >>> with timeit():
        >>>     fun()
    usage 2 :
        >>> ti = timeit()
        # run your code
        >>> print ti()
        
    P.S `with timeit(0)` will convenient stop print 
    '''
    def __init__(self,name='timeit'):
        self.last = self.begin = time.time()
        self.log = isstr(name) or bool(name)
        self.key = name
        self.log_str = name
        if callable(name):
            if "__qualname__" in dir(name):
                self.log_str = name.__qualname__
            elif "__name__" in dir(name):
                self.log_str = name.__name__

    def __call__(self):
        '''返回时间差'''
        t = time.time()
        r = t -self.last
        self.last = t
        return r
    def __enter__(self):
        return self
    def __exit__(self, typee, value, traceback):
        self.t = time.time()-self.begin
        self.p
    @property
    def s(self):
        from .toolLog import strnum
        if 't' not in self.__dict__:
            self.t = self()
        return strnum(self.t,6)
    def __str__(self):
        s='\x1b[36m"%s" spend time: %s\x1b[0m'%(self.log_str, self.s)
        return s
    @property
    def p(self):
        '''直接打印出来'''
        if self.log:
            print(self)


def heatmap(pathOrCode):
    '''show the heatmap of code or python file
    if raise UnicodeDecodeError in Python 2 which may cause by Chinese, Japaneses
    then will replace all symbol not belong ascii to "?$"
    
    Parameters
    ----------
    pathOrCode : str
        .py file path or Python code
    
    Chinese:
    会让代码里面的中文全部失效
    
    Parameters
    ----------
    pathOrCode : str of code or path of .py
        .py文件路径或着python代码
    '''
    beforImportPlt()
    try:
        from pyheat import PyHeat
    except ModuleNotFoundError as e:
        print("Please pip install py-heat!")
        raise e
    import matplotlib.pyplot as plt
    tmppath = 'code-tmp-pyheat-boxx.py'
    
    ispath = pathOrCode.endswith('.py')
    path = pathOrCode if ispath else tmppath
    try :
        if not ispath:
            with open(tmppath,'w') as f:
                f.write(pathOrCode)
        ph = PyHeat(path)
        ph.create_heatmap()
        ph.show_heatmap()
    except UnicodeDecodeError:
        plt.show()
        msg = '''UnicodeDecodeError! try to replace not ascii symbol to '$?' and retry'''
        from boxx import warn
        warn(msg)
        
        with open(path) as f:
            code = f.read()
        code = code.decode('ascii','replace').replace('\ufffd','$?')
        with open(tmppath,'w') as f:
            f.write(code.encode('utf-8'))
        
        ph = PyHeat(tmppath)
        ph.create_heatmap()
        ph.show_heatmap()
    finally:
        if os.path.isfile(tmppath):
            os.remove(tmppath)

def strIsInt(s):
    '''判断字符串是不是整数型'''
    s = s.replace(' ','')
    return s.isdigit() or (s[0]==('-') and s[1:].isdigit())

def strIsFloat(s):
    '''判断字符串是不是浮点'''
    try:
        float(s)
        return True
    except:
        return False

def strToNum(s):
    ''' 若字符串是float or int 是则返回 数字 否则返回本身'''
    if strIsInt(s):
        return int(s)
    if strIsFloat(s):
        return float(s)
    return s

def getArgvDic(argvTest=None):
    '''
    将cmd的`python main.py arg1 arg2 --k v --tag`形式的命令行参数转换为(list, dict)
    若v是数字 将自动转换为 int or float, --tag 将表示为 dic[tag]=True
    
    Return
    ----------
    l : list
        去除第一个参数 文件地址外的 第一个 '--'之前的所有参数
    dic : dict
        `--k v` 将以{k: v}形式存放在dic中
        `--tag` 将以{k: True}形式存放在dic中
    '''
    # TODO support --key=value
    from .toolLog import  pred
    argv = sys.argv
    if argvTest:
        argv = argvTest
    l = argv = list(map(strToNum,argv[1:]))
    code = [(isinstance(x,str) 
        and len(x) >2 and x[:2]=='--') for x in argv]
    dic = {}
    if True in code:
        l = argv[:code.index(True)]
        n = len(code)
        for i,s in enumerate(code):
            x = argv[i]
            if int(s):
                k = x.replace('--','')
                if (i<=n-2 and code[i+1]) or i==n-1: # 不带参数
                    dic[k] = True
                else:  # 带参数
                    dic[k] = argv[i+1]
    if len(dic) or len(l):
        pred('command-line arguments are:\n  %s and %s'%(l,dic))
    return l,dic


def softInPath(softName):
    '''
    是否安装命令为softName的软件，即 判断softName 是否在环境变量里面
    '''
    for p in os.environ['PATH'].split(';' if sysi.win else ':'):
        if os.path.isdir(p) and softName in os.listdir(p):
            return True
    return False

def makedirs(dirr, randomDelay=0.0001):
    os.makedirs(dirr, exist_ok=True)
    return dirr

def execmd(cmd, split=False):
    '''
    execuld cmd and reutrn str(stdout)
    '''
    with os.popen(cmd) as stream:
        if not py2:
            stream = stream._stream
        s = stream.read()
    if split:
        return s.strip().split('\n')
    return s

def addPathToSys(_file_, pathToJoin='.'):
    '''
    将 join(__file__, pathToJoin)  加入 sys.path

    Parameters
    ----------
    _file_ : str
        .py 文件的路径 即__file__ 变量
    pathToJoin : str, default '.'
        相对路径
    '''
    from os.path import abspath,join,dirname
    apath = abspath(join(dirname(abspath(_file_)),pathToJoin))
    if apath not in sys.path:
        sys.path.append(apath)
    return apath

def getFatherFrames(frame=0, endByMain=True):
    '''
    返还 frame 的所有的父 frame 即 Call Stack
    
    Parameters
    ----------
    frame : frame or int, default 0
        if int:相对于调用此函数frame的 int 深度的对应frame
    endByMain : bool, default True
        为 True 则在第一个 frame.f_locals['__name__'] == '__main__' 处停止搜寻
        目的是去除 IPython 自身多余的 Call Stack
    '''
    if frame is False and endByMain:
        frame, endByMain = 0, False
    if isinstance(frame, int):
        frame = sys._getframe(1 + frame)
    fs = []
    while frame:
        fs.append(frame)
        if endByMain:
            if '__name__' in frame.f_locals and frame.f_locals['__name__'] == '__main__':
                break
        frame = frame.f_back
    return fs



mainFrame = []
def getMainFrame(frame=0):
    '''
    return a main frame from father frames that first frame.f_locals['__name__'] == '__main__' 
    
    Parameters
    ----------
    frame : frame or int, default 0
        if int:相对于调用此函数frame的 int 深度的对应frame
    '''
    if len(mainFrame):
        return mainFrame[0]
    fs = getFatherFrames(frame=frame+1, endByMain=True)
    main = fs[-1]
    mainFrame.append(main)
    return main
getMainFrame()


rootFrame = []
def getRootFrame():
    '''
    return interactive frame
    '''
    if len(rootFrame):
        return rootFrame[0]
    frame=0
    endByMain=False
    fs = getFatherFrames(frame=frame+1, endByMain=endByMain)
    root = getMainFrame()
    
    for f in fs:
        if f.f_code.co_filename.startswith('<ipython-input-'):
            root = f
            break
    rootFrame.append(root)
    return root
getRootFrame()

if __name__ == "__main__":

    pass


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
sysc.py: system config
"""
from __future__ import unicode_literals
import numpy as np
from .. import *
from ..ylsys import cpun, cloud, cuda, usecuda
from ..ylimg import npa, r
from ..tool import FunAddMagicMethod, nextiter, withfun, pred

from ..ylcompat import py2


from collections import OrderedDict
from functools import wraps
import matplotlib.pyplot as plt
import skimage.data as sda


from torch.autograd import Variable
import torch.utils.data

th = torch
nn = th.nn
from torch.nn import (
    Conv2d,
    Linear,
    ConvTranspose2d,
    BatchNorm2d,
    ReLU,
    Tanh,
    Softmax2d,
    CrossEntropyLoss,
    DataParallel,
    MSELoss,
    MaxPool2d,
    AvgPool2d,
    Module,
    functional,
    Sequential,
)

Tensor = torch.Tensor
F = functional

import torchvision
import torchvision.transforms as transforms
import torchvision.datasets as datasets


# add summary to torch.nn.Module
nn.Module.summary = lambda self, inputShape=None, group=None, ganNoise=False: __import__(
    "torchsummary"
).summary(
    self,
    inputShape or getModelDefaultInputShape(self, group, ganNoise),
    device=["cuda", "cpu"]["cpu" in str(getpara(self).device)],
)


def dedp(model):
    """get raw model instead of torch.nn.DataParallel """
    return model.module if isinstance(model, torch.nn.DataParallel) else model


if "Module" in str(torch.nn.Module.load_state_dict):
    torch.rawModule = rawModule = torch.nn.Module.load_state_dict
else:
    rawModule = torch.rawModule


def tryLoad(self, state_dict, strict=True):
    try:
        return rawModule(self, state_dict, strict)
    except (KeyError, RuntimeError):
        print(
            "\x1b[31m%s\x1b[0m"
            % '\n"try strict=False! in Module.load_state_dict() " messge from boxx.ylth \n'
        )
        para = state_dict
        para = OrderedDict([(k.replace("module.", ""), v) for k, v in para.items()])
        return rawModule(self, para, strict)


rawDataLoader = th.utils.data.DataLoader


try:
    from typing import TypeVar, Generic
    
    T_co = TypeVar('T_co', covariant=True)
    
    class DataLoaderForCPU(rawDataLoader, Generic[T_co]):
        def __init__(self, *l, **kv):
            rawDataLoader.__init__(self, *l, **kv)
            self.pin_memory = False
except:
    class DataLoaderForCPU(rawDataLoader):
        def __init__(self, *l, **kv):
            rawDataLoader.__init__(self, *l, **kv)
            self.pin_memory = False

def toCpu():
    cudaAttri = lambda self, *l, **kv: self
    nn.Module.cuda = cudaAttri
    Variable.cuda = cudaAttri
    torch.Tensor.cuda = cudaAttri
    rawTo = torch.Tensor.to
    def to(self, device=None, *args, **kwargs):
        if isinstance(device, (str, torch.device)):
            device = "cpu"
        return rawTo(self, device, *args, **kwargs)
        
    torch.Tensor.to = to
    torch.Tensor.is_cuda = True

    #    class FakeDataParallel(torch.nn.DataParallel):
    #        def __init__(self, x):
    #            super(FakeDataParallel, self).__init__()
    torch.nn.DataParallel = cudaAttri

    class withh:
        def __init__(self, *l):
            pass

        def __call__(self):
            return self

        def __enter__(self):
            return self

        def __exit__(self, typee, value, traceback):
            pass

    torch.cuda.device = withh
    from boxx import fnone

    torch.cuda.set_device = fnone
    torch.cuda.is_available = lambda: True
    torch.cuda.is_bf16_supported = lambda : False
    torch.cuda.device_count = lambda *l: 0

    th.cuda.LongTensor = th.LongTensor
    th.cuda.DoubleTensor = th.DoubleTensor
    th.cuda.IntTensor = th.IntTensor
    th.cuda.ShortTensor = th.ShortTensor
    th.cuda.ByteTensor = th.ByteTensor
    th.cuda.HalfTensor = th.HalfTensor
    th.cuda.CharTensor = th.CharTensor
    th.cuda.FloatTensor = th.FloatTensor

    #    def warp(f):
    #        @wraps(f)
    #        def DataLoader(*l, **kv):
    #            if 'pin_memory' in kv:
    #                kv.pop('pin_memory')
    #            r = f(*l, **kv)
    #            return r
    #        return DataLoader


    th.utils.data.DataLoader = DataLoaderForCPU

    rawThLoad = torch.load

    def torchLoad(*l, **kv):
        return rawThLoad(*l, **(kv.update({"map_location": "cpu"}) or kv))

    torch.load = torchLoad
    
    raw_load_state_dict = nn.Module.load_state_dict
    def load_state_dict(self, state_dict, *l,
                        strict= True, **kv):
        startswith_module_rate = lambda dic:sum([key.startswith("module.") for key in dic])/len(dic)
        
        
        if (startswith_module_rate(state_dict) == 1 and 
            startswith_module_rate(self.state_dict()) < .25):
            state_dict = {k.replace("module.",""):v for k, v in state_dict.items()}
        return raw_load_state_dict(self, state_dict, *l,
                            strict=strict, **kv)
            
            
    nn.Module.load_state_dict = tryLoad
    torch.nn.modules.module.Module.load_state_dict = load_state_dict


usecpu = (not cuda and usecuda == "auto") or not usecuda
if usecpu:
    toCpu()
_TensorBase = torch._TensorBase if "_TensorBase" in dir(torch) else torch._C._TensorBase


def tht(t):
    """
    anything t to torch.Tensor
    """
    if not isinstance(t, _TensorBase):
        t = th.from_numpy(npa - t).cuda()
    return t.cuda()


tht = FunAddMagicMethod(tht)

# t = tht(r).float()
t = th.from_numpy(r).float()
if cuda:
    t = t.cuda()


def recursive_func(func):
    def inner_func(batch, device=None):
        if isinstance(batch, dict):
            return {k: inner_func(v) for k, v in batch.items()}
        elif isinstance(batch, (list, tuple)):
            return type(batch)([inner_func(x) for x in batch])
        else:
            return func(batch, device=device)

    return inner_func


def to_tensor(x, device=None):
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    if isinstance(x, torch.Tensor):
        x = x.cuda() if device is None else x.to(device)
    return x


batch_to_tensor = recursive_func(to_tensor)


def to_numpy(x, device=None):
    if isinstance(x, torch.Tensor):
        x = x.cpu().numpy()
    return x


batch_to_numpy = recursive_func(to_numpy)


def batchToTensor(batch, device=None):
    """
    turn a dataloader's batch to tensor (from dpflow).
    support dict, list, tuple as a batch
    """

    def to_tensor(x):
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x)
        if isinstance(x, torch.Tensor):
            x = x.cuda() if device is None else x.to(device)
        return x

    if isinstance(batch, dict):
        return {k: to_tensor(v) for k, v in batch.items()}
    if isinstance(batch, (list, tuple)):
        return [to_tensor(v) for v in batch]


def batchToNumpy(batch):
    """
    turn a dataloader's batch to numpy (for dpflow).
    support dict, list, tuple as a batch
    """
    if isinstance(batch, dict):
        return {
            k: (v.cpu()).numpy() if isinstance(v, torch.Tensor) else v
            for k, v in batch.items()
        }
    if isinstance(batch, (list, tuple)):
        return [(v.cpu()).numpy() if isinstance(v, torch.Tensor) else v for v in batch]


@wraps(torch.autograd.Variable)
def var(t, *l, **kv):
    t = tht(t)
    t = Variable(t, *l, **kv)
    return t.cuda()


var = FunAddMagicMethod(var)


def kaimingInit(model):
    stateDict = model.state_dict()
    for key in stateDict:
        tag = 0
        t = stateDict[key]

        if (t.ndimension() == 4 and t.shape[-1] > 2) or "conv" in key:
            nn.init.kaiming_normal(stateDict[key], mode="fan_out")
            tag = 1
        elif "bn" in key and "weight" in key:
            stateDict[key][...] = 1
            tag = 1
        elif "bias" in key:
            stateDict[key][...] = 0
            tag = 1
        elif "fc" in key and t.ndimension() == 2:
            nn.init.kaiming_normal(t)


def getModelDefaultInputShape(model, group=None, ganNoise=False):
    para = nextiter(model.parameters())
    shape = para.shape
    if len(shape) == 4 and shape[-1] > 2:
        default = (shape[1], 244, 244)
    elif len(shape) == 4 and shape[-1] == shape[-2] == 1:
        default = (shape[1], 244, 244)
        if ganNoise:
            default = (shape[1], 1, 1)
    elif len(shape) == 2:
        default = (shape[1],)
    if group:
        default = (shape[1] * group, 244, 244)
    return default


def genModelInput(model, inputShape=None, group=None, ganNoise=False, batchn=2):
    """
    Auto generate a Tensor that could as model's input
    
    Usage
    -----
    >>> inp = genModelInput(model)
    >>> result = model(inp)
    >>> tree(result)
    
    Parameters
    ----------
    model: nn.Module
        nn.Module
    inputShape: tuple or list, default None
        By default, inputShape will auto calculate through model's first parameters 
    group: int, default None
        If the first conv has is a group conv, please provide the group num
    ganNoise: bool, default False
        set True, if input is a 1dim vector of noise for GAN
    batchn: int, default 2
        batch number, consider the BatchNorma opr, default of batchn is 2
    """
    inputShape = inputShape or getModelDefaultInputShape(model, group, ganNoise)
    para = getpara(model)
    inp = th.rand((batchn,) + inputShape, dtype=para.dtype, device=para.device)
    if not ganNoise:
        from skimage.data import astronaut

        img = astronaut().mean(-1) / 255.0
        mean = img.mean()
        std = ((img - mean) ** 2).mean() ** 0.5
        normaed = (img - mean) / std
        feat = nn.functional.interpolate(
            tht - [[normaed]], inputShape[-2:], mode="bilinear",
        )
        inp[:] = feat.to(para.device)
    return inp


class HookRegister:
    def __init__(self, module, hook, direct="f"):
        self.hook = hook
        self.hooks = hooks = []
        self.module = module

        def apply(module):
            if direct == "f":
                hooks.append(module.register_forward_hook(hook))
            elif direct == "b":
                hooks.append(module.register_backward_hook(hook))

        module.apply(apply)

    def remove(self):
        for h in self.hooks:
            h.remove()

    def __enter__(self):
        return self

    def __exit__(self, *l):
        self.remove()


def removeAllHook(module):
    def apply(module):
        module._forward_hooks = OrderedDict()

    module.apply(apply)


from boxx import log, ylimgTool, g

ar = FunAddMagicMethod(lambda x: log - ylimgTool.prettyArray(x))


def nanDete(t, globalg=False):
    nan = th.isnan(t).sum()
    if nan:
        if globalg:
            g(1)
        ar(t)
        raise LookupError("Has torch.nan")


def getpara(m):
    """get first parameter"""
    return nextiter(m.parameters())


def getpara0(m):
    return getpara(m).view(-1)[0]


def getgrad(m):
    return getpara(m).grad


def getgrad0(m):
    grad = getpara(m).grad
    return None if grad is None else grad.view(-1)[0]


getpara, getpara0, getgrad, getgrad0 = map(
    FunAddMagicMethod, [getpara, getpara0, getgrad, getgrad0]
)


def get_loss(model_output):
    if isinstance(model_output, (list, tuple)):
        loss = sum([i.sum() for i in model_output])
        return loss
    if isinstance(model_output, dict):
        loss = sum([i.sum() for i in model_output.values()])
        return loss
    return model_output.sum()


def vizmodel(m, inputShape=None, group=None, ganNoise=False, batchn=2, output=""):
    x = genModelInput(
        m, inputShape=inputShape, group=group, ganNoise=ganNoise, batchn=batchn
    )
    from torchviz import make_dot

    x.to(getpara(m))
    model_output = m(x)
    loss = get_loss(model_output)
    graph = make_dot(loss, params=dict(m.named_parameters()))
    if output:
        graph.render(output, format="png")
    return graph


def flatten(t, dim=-1):
    """
    >>> t = shape(1,2,3,4) 
    >>> flatten(t, dim=-2)
    shape(1,6,4)
    """
    shape = list(t.shape)
    shape[dim - 1] *= shape[dim]
    shape.pop(dim)
    return t.reshape(tuple(shape))


def hasnan(tensor):
    return not bool(torch.isfinite(tensor).all())


def pthnan(pth="/home/dl/github/maskrcnn/output/mix_11/model_final.pth"):
    dic = torch.load(pth)
    from collections import OrderedDict

    def getOrderedDict(seq):
        if isinstance(seq, OrderedDict):
            return seq
        if isinstance(seq, dict):
            seq = list(seq.values())
        if not isinstance(seq, (tuple, list)):
            return None
        for s in seq:
            re = getOrderedDict(s)
            if re is not None:
                return re

    od = getOrderedDict(dic)
    tensor = list(od.values())[-1]
    print('"%s"has nan: %s' % (tensor[..., :10], hasnan(tensor)))


if __name__ == "__main__":
    l = [
        "LongTensor",
        "DoubleTensor",
        "IntTensor",
        "ShortTensor",
        "ByteTensor",
        "HalfTensor",
        "CharTensor",
        "FloatTensor",
    ]
    formatt = "th.cuda.%s = th.%s"
    print("\n".join([formatt % (i, i) for i in l]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed May  9 00:55:51 2018

@author: leiYang
"""
from __future__ import unicode_literals, print_function

#what-
from boxx import *
import boxx
test = True
test = False
if test:
    from toolTest import *
    
    from ylimgTest import *
    
    from ylmlTest import *

pattern = 'TODO'
findinRoot(pattern,root='..')
#findinRoot(up=pattern,root='..')
#replaceAllInRoot(pattern,'Function',root='..')


if __name__ == '__main__0':
    pass
    from boxx.ylth import *
    
    def f(self,*l, **kv):
        init = self.__init__
        code = init.__code__
        ca = code.co_argcount
        cv = code.co_varnames
#        type(self)()
        
#        log-prettyFrameStack()
        frames = getFatherFrames(1)
        for f in frames:
            log-f.f_code.co_name
            if f.f_code.co_name != '__init__':
                break
            loc = f.f_locals
#        tree-loc
        args = {k:loc[k] for k in cv}
        tree-args
        rawModuleInit(self)
#        out()
#        g.f = frames = getFatherFrames()
#        p/map2(lambda f:f.f_locals.get('self'),frames[::])
#        tree-locals()
    nn.Module.__init__
    
#    class nnModule(nn.Module):
#        def __init__(self, *l, **kv):
#            log-prettyFrameStack()
    if '__init__' in str(nn.Module.__init__):
        rawModuleInit = nn.Module.__init__
    nn.Module.__init__ = f
#        nn.Module = nnModule
#    conv = nn.Conv2d(1,2,(3,3))
    class Mo(nn.Module):
        def __init__(self,b='b',**l):
            super(Mo, self).__init__()
    a= Mo('a')
    
#    what- a.__init__.__code__


    fs=g.f





# -*- coding: utf-8 -*-
"""
tools for pandas 

@author: yl
"""

from __future__ import unicode_literals
import pandas as pd

def df2dicts(df, inculdeIndex=False):
    '''
    input a df, return to a list of dict
    
    Parameters
    ----------
    df : DataFrame
        pandas.DataFrame
    inculdeIndex : bool, default False
        return dic include index
    '''
    dicts = [dict(row) for index,row in df.iterrows()]
    if inculdeIndex:
        for index, dic in zip(df.index, dicts):
            dic['index'] = index
    return dicts

if __name__ == '__main__':
    pass
    df = pd.DataFrame({
                       0:list(range(5)),
                       1:list(range(10,15)),
                       'a':list("abcde"),
                       })
    df.set_index(0,inplace=True)




# -*- coding: utf-8 -*-
"""
tools for Image Processing and Visualization

@author: yanglei
"""
from .ylimgTool import sk, np, plot, sda
from .ylimgTool import (show, showb, shows, loga, tree, treem, treea, dira, what, wtf,
                        tprgb, torgb, frombgr, mapp, normalizing, norma, imsave, imread, npa, 
                       getShapes, greyToRgb, warpResize, resize, histEqualize, StructLogFuns)
from .ylimgTool import (uint8, padding, toPng, randomm , r, labelToColor,getDefaultColorList,
                       standImg, getMeanStd, boolToIndex, discribArray, prettyArray)

#from .ylimgVideoAndGif import gifSave, videoToImgs, base64Img, generateBigImgForPaper
from . import ylimgVideoAndGif 


if __name__ == '__main__':

    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals


class dictToObj(dict):
    '''
    将dic转换为可以用dic模式访问的obj
    '''
    def __init__(self, dic):
        dict.__init__(self)
#        self.dic = dic
        for k,v in list(dic.items()):
            setattr(self,k,v)
    def __getitem__(self,k):
        return getattr(self,k)
    def __setitem__(self,k,v):
        return setattr(self,k,v)









# -*- coding: utf-8 -*-
from .tool import *
from .ylimg import *
from .ylml import *
from .ylnp import *
from . import tool
from . import ylimg
from . import ylml
from . import ylnp


if __name__ == '__main__':
    
    pass




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
sample a small data set.
For each types files in dir, only copy 50 files to target.
won't copy size > 512MB file

Created on Sat Jun 30 14:58:44 2018

@author: yanglei
"""

import argparse, os
import shutil
from collections import defaultdict

from os.path import join, isdir, basename
pathjoin = join

parser = argparse.ArgumentParser(description="""
sample a small data set.
For each types files in dir, only copy 50 files to target.
won't copy size > 512MB file""")
parser.add_argument('source_target', nargs='*', default=None, type=str)
parser.add_argument('-n', default=10, type=float, 
                    help='How many data to copy')
parser.add_argument('-x', '--max', default=50, type=int, 
                    help='max numbers of the file if number > max: only copy top n file ')

parser.add_argument('-s', '--max_size', default=512, type=int, 
                    help="max size of the file if size > max_size won't copy, The unit is MB")


#parser.add_argument('source', nargs='*', default='boxx', type=str)
#parser.add_argument('target', nargs='0', default='/tmp/', type=str)


def path2type(path):
    if isdir(path):
        return "/dir"
    try:
        ind = path.rindex('.')
    except ValueError:
        return "/~."
    return path[ind+1:]

def filterFileList(l):
    if len(l) < arg.max:
        return l    
    n = arg.n
    if n < 1:
        n = len(l) * n
    n = int(n)
    l = sorted(l)
    return l[:n] 

def copyfile(src, dst):
    size = os.path.getsize(src)/float(1024**2)
    if size > arg.max_size:
        new = '%s_%sMB.fake'%(dst, int(size))
        with open(new ,'w',) as f:
            f.write('size = %s MB'%size)
        return 
    shutil.copyfile(src, dst)
    
def listdirWithFun(source='.', target='/tmp', __first=True):
    '''对source路径及子路径下的每个path 执行fun
    
    Parameters
    ----------
    source : str, default '.'
        路径
    fun : function
        对每个子路径执行fun(path)
    '''
    if __first:
        source = os.path.abspath(source)
        if isdir(target):
            target = pathjoin(target, basename(source))
        if not isdir(target):
            os.makedirs(target, exist_ok=True)    
        
#    paths = [pathjoin(source, p) for p in os.listdir(source)]
    paths = os.listdir(source)
    
    d = defaultdict(lambda :[])
    [d[path2type(pathjoin(source, p))].append(p) for p in paths]
    
    d = {k:filterFileList(v) for k,v in d.items()}
    
    dirs = d.pop('/dir') if '/dir' in d else []
    
    [[copyfile(pathjoin(source, p), pathjoin(target, p))  for p in l] for l in d.values()]
    
    for dirr in dirs:
        targetd = pathjoin(target, dirr)
        if not isdir(targetd):
            os.makedirs(targetd, exist_ok=True)
        listdirWithFun(pathjoin(source, dirr), target=targetd, __first=False)

if __name__ == '__main__':
    arg = parser.parse_args()
    arg.max = max(arg.max, arg.n)
    arg.source = arg.source_target[0] if len(arg.source_target) else '.'
    arg.target = arg.source_target[1] if len(arg.source_target)>=2 else '/tmp/'
    print(arg)
    listdirWithFun(arg.source, arg.target)








# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os, sys
import warnings
from functools import wraps

from ..ylsys import py2, sysi

class BoxxException(Exception):
    '''
    root Exception for boxx
    '''
    pass

class BoxxWarning(Warning):
    '''
    root warninng for boxx
    '''
    pass

class OffScreenWarning(BoxxWarning):
    pass

class Except():
    '''
    get traceback frame in with 
    
    >>> with excep:
    >>>     1/0
    >>> dira(excep)
    '''
    def __init__(self, deep=0):
        self.deep = deep
    def __enter__(self):
        pass
    def __exit__(self, typee, value, traceback):
        deep = self.deep
        while deep:
            deep -= 1
            traceback = traceback.tb_next
        self.type = typee
        self.value = self.v = value
        self.traceback = self.t = traceback
        self.frame = self.f = traceback.tb_frame
excep = Except()


def getExcept(fun):
    '''
    exec `fun()` and return (Exception, trace, frame)
    '''
    try:
        exc = Except(1) 
        with exc:
            fun()
    except Exception as ee:
        e = ee
        return e, exc.traceback, exc.frame
    
def warn(msg, warnType=BoxxWarning, filename=None, line=None, module='boxx', blue=False):
    '''
    log a warning of type warnType warn will auto fill filename and line 
    '''
    msg = '''%s
    %s'''%(('\x1b[36m%s\x1b[0m' if blue else '%s')% 'warning from boxx', msg)
    if filename is None or line is None:
        f = sys._getframe(1)
        c = f.f_code
        filename = c.co_filename if filename is None else filename
        line = c.co_firstlineno if line is None else line 
    warnings.warn_explicit(msg, warnType, filename, line, module)

warn1timeCache = {}
@wraps(warn)
def warn1time(msg, *l, **kv):
    '''
    log a warning of type warnType warn will auto fill filename and line 
    
    warn only one time
    '''
    if not warn1timeCache.get(msg):
        warn(msg, *l, **kv)
        warn1timeCache[msg] = True
    
getsize = os.path.getsize

def getsizem(path='.'):
    '''
    返回 path 的大小 支持文件夹 单位为 MB
    '''
    if os.path.isdir(path):
        return sum([getsizem(os.path.join(path, p)) for p in os.listdir(path)])
    return os.path.getsize(path)/float(1024**2)


def fileJoinPath(_file_,path='.'):
    '''
    返回 __file__ + 相对路径 path 后的绝对路径
    '''
    from os.path import abspath,join,dirname
    apath = abspath(join(dirname(abspath(_file_)),path))
    return apath

def filename(path):
    '''
    将路径和后缀名除去 只留下文件名字
    '''
    filen = name = os.path.basename(path)
    if '.' in name:
        filen = name[:name.rindex('.')]
    return filen

def relfile(relative_path):
    '''
    Return a absolute version of a relative_path relative the __file__
    '''
    frame = sys._getframe(1)
    if '__file__' not in frame.f_globals:
        return relative_path
    _file_ = frame.f_globals['__file__']
    abspath = os.path.abspath(os.path.join(_file_, '..', relative_path))
    return abspath

def listdir(path=None):
    path = path or '.'
    return os.listdir(path)

def openread(path, encoding='utf-8'):
    '''
    返回path文件的文本内容
    '''
    if py2:
        with open(path, 'r') as f:
            return f.read()
    with open(path, 'r', encoding=encoding) as f:
        strr = f.read()
    return strr
def openwrite(strr, path, mode='w', encoding='utf-8'):
    '''
    将strr写入path
    '''
    if py2:
        with open(path, mode) as f:
            f.write(strr)
        return path
    with open(path, mode, encoding=encoding) as f:
        f.write(strr)
    return path


def validFilename(filename, replaceBy='_'):
    '''
    return validate filename 
    '''
    import re
    if sysi.win:
        rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
    else:
        rstr = r"[\/]" # ' / '
    newName = re.sub(rstr, replaceBy, filename)
    return newName

def first_exist_dir(*dirs):
    """Input dirs and return the first exist dir.
    If none dir exist, return First
    """
    if len(dirs) == 1 and isinstance(dirs[0], (list, tuple)):
        dirs = dirs[0]
    for dirr in dirs:
        if os.path.isdir(dirr):
            return dirr
    return dirs[0]

def loadjson(path):
    import json
    with open(path, 'r') as f:
        js = json.load(f)
    return js

def savejson(obj, path=None, indent=None):
    import json
    import numpy as np
    
    class NumpyEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, (np.ndarray, np.number)):
                return obj.tolist()
            return json.JSONEncoder.default(self, obj)
    if path is None:
        return json.dumps(obj, indent=indent, cls=NumpyEncoder).encode("utf-8")
    with open(path, 'w') as f:
        json.dump(obj, f, indent=indent, cls=NumpyEncoder)
    return path 
    
def replaceTabInPy(dirr='.'):
    '''
    将所有tab换成4个空格
    '''
    from glob import glob
    from .toolLog import log
    pys = glob(os.path.join(dirr, '*.py'))
    for py in pys:
        code = openread(py)
        log(py,code.count('\t'))
        new = code.replace('\t',' '*4)
        openwrite(new, py)

def saveData(data, name='pickle_of_boxx', log=False):  #保存进度
    '''
    保存二进制数据
    '''
    import pickle
    if log:
        print('正在将数据写入',os.path.abspath('.'),'下的文件:“'+name+'”，请稍等。。。')
    with open(name, "wb") as f:
        pickle.dump(data,f)
    if log:
        print('\n文件“'+name+'”已保存在',os.path.abspath('.'),'目录下!')

def loadData(name='pickle_of_boxx', log=False):  #载入数据
    import pickle
    if not os.path.isfile(name):
        print('在',os.path.abspath('.'),'目录下,“'+name+'”文件不存在，操作失败！')
    if log:
        print('正在读取',os.path.abspath('.'),'目录下的文件:“'+name+'”\n请稍等。。。')
    with open(name,"rb") as f:
        data = pickle.load(f)
    f.close()
    if log:
        print('文件:“'+name+'”读取成功！')
    return data

def browserOpen(url):
    '''
    open url with browser
    if can't open browser raise warn
    '''
    import webbrowser
    if not webbrowser.open_new_tab(url):
        from boxx import warn
        warn('''can't open url with web browser, plaese open url:"%s" in your browser'''%url)
        
if __name__ == "__main__":

    pass


# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import re
import os
import random
from os.path import join,isdir,isfile
from glob import glob
from collections import defaultdict

from .toolIo import openread, openwrite, getsizem
from .toolLog import log

from ..ylsys import py2

intround = lambda floatt: int(round(floatt))

__increase_recording = defaultdict(lambda:-1)
def increase(namespace=None):
    '''
    从0开始 每调用一次返回的数值自增 1， 类似 SQL 中的 AUTO_INCREMENT 字段
    
    Parameters
    ----------
    namespace : hashable, default None
        用于计数的命名空间
    '''
    __increase_recording[namespace] += 1
    return __increase_recording[namespace]
increase.d = __increase_recording
def filterList(key, strs):
    '''
    对一个str列表 找出其中存在字符串 key的所有元素
    '''
    return list(filter((lambda strr: key in strr),strs))

def findints(strr):
    '''
    返回字符串或字符串列表中的所有整数 ,r"\d+\d*"
    '''
    if isinstance(strr,(list,tuple)):
        return list(map(findints, strr))
    return list(map(int,re.findall(r"-?\d+\d*",strr)))

def randint(maxx=100):
    return random.randint(0, maxx)

def randfloat():
    return random.random()

def randchoice(seq, num=None):
    '''
    随机选择一个列表内的一个或num个元素
    '''
    if num is None:
        return random.choice(seq)
    return random.sample(seq, num)

def listdirWithFun(root='.',fun=None):
    '''对root路径及子路径下的每个path 执行fun
    
    Parameters
    ----------
    root : str, default '.'
        路径
    fun : function
        对每个子路径执行fun(path)
    '''
    paths = glob(join(root,'*'))
    for path in paths:
        if isdir(path):
            listdirWithFun(path,fun)
        if fun :
            fun(path)

def replaceAllInRoot(old, new, root='.', types='py'):
    '''对root路径及子路径下的每个types文件类型 执行替换old到new的操作 
    ps. old,new 应尽可能的长 包含更多的上下文信息 以避免错误的替换
    
    Parameters
    ----------
    old : str
        被替换部分
    new : str
        替换的新的内容
    root : str, default '.'
        路径 即根文件夹
    types : str or list, default 'py'
        需要被替换的文件类型
        str:单个类型, list:多个类型 默认为 'py'文件
    '''
    if not isinstance(types,(list,tuple)):
        types = [types]
    def replace(path):
        if isfile(path):
            if '.' in path and path.split('.')[-1] in types:
                code = openread(path)
                if old in code:
                    log('replaceAllInRoot: %s  old=%s  new=%s'%(path, old, new))
                    openwrite(code.replace(old,new),path)
    listdirWithFun(root, replace)

def findinRoot(pattern='', root='.', maxsize=1, types=None, var=None, up=None, re=None, 
               exclude=None):
    '''
    在root及子路径下的所有文件中 查找 pattern。存在 则打印出对应文件的那一行。
    
    Parameters
    ----------
    pattern : str
        要查找的内容
    root : str, default '.'
        路径 即根文件夹
    maxsize : number, default 1
        被查找文件的最大文件大小，单位为 MB
        为避免不小心查找太大的二进制文件，默认不超过 1MB
        设为 None，则不设大小限制
    types : str or list, default None
        被查找文件的类型限制
        str:单个类型, list:多个类型 默认为所有文件
    var : str or bool, default None
        use re.compile('(^|[^a-zA-Z0-9_])%s([^a-zA-Z0-9_]|$)'%var) to find var name
    up : str or bool, default None
        Ignoring letter case of variable names
    re : str or bool, default None
        use re.compile(re) to search each line
    exclude : list , default ['pyc', 'swp', 'swo',]
        exclude file types in this list
    '''
    from re import compile as recompile
    if types is not None :
        if not isinstance(types,(list,tuple)):
            types = [types]
        types = [t.replace('.', '') for t in types]
    if exclude is None:
        exclude = []
    exclude += ['pyc', 'swp', 'swo', 'gz', 'whl', 'jpg', 'png', 'npz', 'bin']
    def intypes(path):
        typee = '.' in path and path.split('.')[-1].lower()
        if types is None:
            return typee not in exclude
        return typee in types
    if not pattern:
        pattern = var or up or re
    searchin = lambda strr: pattern in strr
    if var:
        pa = '''(^|[^a-zA-Z0-9_'"])%s([^a-zA-Z0-9_'"]|$)'''%pattern
        pa = recompile(pa)
        searchin = lambda strr: pa.search(strr)
    elif up:
        lower = pattern.lower()
        searchin = lambda strr: lower in strr.lower()
    elif re:
        pa = recompile(pattern)
        searchin = lambda strr: pa.search(strr)
    def find(path):
        if isfile(path):
            if (getsizem(path) <= maxsize or maxsize is None) and intypes(path):
                try:
                    s = openread(path)
                    if py2:
                        from .toolLog import tounicode
                        s = tounicode(s)
                    lines = [(i, l) for i,l in enumerate(s.split('\n'), 1) if searchin(l)]
                    if not len(lines):
                        return 
                    print('"%s" in "%s" with %s'%('\x1b[31m%s\x1b[0m'%pattern, '\x1b[36m%s\x1b[0m'%path, '\x1b[35m%s Lines\x1b[0m'%len(lines)))
                    for (i, l) in lines:
                        if l.startswith('   '):
                            l = '... '+l.strip()
                        if len(l) > 83:
                            l = l[:80] +' ...'
                        print('\t%s:%s'%('\x1b[35m%s\x1b[0m'%i, '\x1b[31m%s\x1b[0m'%l))
                    print("")
                except:
                    return 
    listdirWithFun(root, find)

def iscn(char):
    '''
    Does a char is chinese? 
    '''
    if u'\u4e00' <= char <= u'\u9fff':
        return True
    return False
    

def zipTar(paths, tarp):
    '''
    tar file for ,especially for label++
    '''
    import tarfile  
    with tarfile.open(tarp,'w') as tar:
        for path in paths:
            tar.add(path, arcname=os.path.basename(path))

def camel2snake(variable_name):
    s = variable_name[0].lower()
    for c in variable_name[1:]:
        if c.isupper():
            s += '_' + c.lower()
        else:
            s += c
    return s

def snake2camel(variable_name, lower_camel_case=False):
    s = variable_name[0].lower() if lower_camel_case else variable_name[0].upper()
    for idx, c in enumerate(variable_name[1:]):
        if variable_name[idx] == '_':
            s += c.upper()
        elif c != '_':
            s += c
    return s

if __name__ == "__main__":
     
    string=["A001.45，b5，6.45，8.82",'sd4 dfg77']
    print(findints(string))
    pass


# -*- coding: utf-8 -*-
"""
General Python Tools for Debugging, Logging, Magic .etc

@author: yanglei
"""

from __future__ import unicode_literals

import os, sys, time
from ..ylsys import py2

from .toolLog import (stdout, log, logc, printt, PrintStrCollect, printToStr, logg, tounicode, tostrpy2,
                      shortDiscrib, shortStr, discrib, getDoc, tabstr, ignoreWarning, LogException, 
                      LogLoopTime, SuperG, sg, g, mg, gg, cf, p, pp, lc, out)
from .toolLog import colorFormat, clf, pblue, pred, pdanger, perr, pinfo, decolor
from .toolLog import strnum, percentStr, notationScientifique
from .toolLog import prettyClassFathers, prettyFrameLocation, prettyFrameStack, wp, wg, wgg
from .toolLog import localTimeStr, gmtTimeStr, time_str_to_stamp, timeGap, timegap, logGap

from .toolStructObj import (dicToObj, dicto, typeNameOf, typestr, strMethodForDiraAttrs, 
                            getfathers, getfather, generator, nextiter, listToBatch, 
                            sliceInt, sliceLimit, ll, FunAddMagicMethod, mf, fnone, addCall,
                            setself, unfoldself, withfun, withattr, isinstancestr)

from .toolIo import (getsize, getsizem, listdir, filename, relfile, openread, openwrite, validFilename,
                    first_exist_dir, saveData, loadData, loadjson, savejson, fileJoinPath,  warn, 
                    replaceTabInPy, warn1time, BoxxException, BoxxWarning, OffScreenWarning,
                    Except, excep, getExcept, browserOpen)
save_data = saveData
load_data = loadData

from .toolSystem import (crun, performance, timeit, heatmap, getArgvDic, 
                         softInPath, makedirs, execmd, addPathToSys)
from .toolSystem import importAllFunCode, impt, inpkg, importByPath, tryImport, FakeModule, removeImportSelf,  removeimp
from .toolSystem import getMainFrame, getRootFrame, getFatherFrames#, exceptionHook

from .toolFunction import (getFunName, FuncArgs, SaveArguments, dynamicWraps, setTimeout, pipe,
                           setInterval, multiThread, mapmp, mapmt, maptry, retry)

from .toolTools import (intround, increase, filterList, findints, randint, randfloat, randchoice,
                       listdirWithFun, replaceAllInRoot, findinRoot, iscn, zipTar,
                       camel2snake, snake2camel)

from .toolMarkdown import Markdown

from .toolGui import ter, nau, vscode

# Compatible with previous `boxx.glob` and `glob module`
import glob as globModule
glob = globModule.glob
glob.__dict__.update(globModule.__dict__)

from os.path import join as pathjoin
from os.path import basename, isfile, isdir, dirname, abspath, expanduser

from collections import namedtuple, defaultdict, Counter, OrderedDict
dictd = defaultdict
odict = OrderedDict

from time import sleep
from functools import reduce, wraps
from operator import add, sub, mul
if py2:
    from operator import div
else :
    from operator import truediv as div

from functools import wraps

if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
import numpy as np
from ..ylimg import imread


from collections import Iterator 
class GenSimg(Iterator):
    '''
    随机生成小图片simg及gt 的迭代器，默认使用1Gb内存作为图片缓存
    默认生成simg总面积≈所有图像总面积时 即结束
    '''
    def __init__(self, imggts, simgShape, handleImgGt=None,
                 batch=1, cache=None,iters=None,
                 timesPerRead=1,infinity=False):
        '''
        imggts: zip(jpgs,pngs)
        simgShape: simg的shape
        handleImgGt: 对输出结果运行handleImgGt(img,gt)处理后再返回
        batch: 每次返回的batch个数
        cache: 缓存图片数目, 默认缓存1Gb的数目
        timesPerRead: 平均每次读的图片使用多少次(不会影响总迭代次数),默认1次
        iters: 固定输出小图片的总数目，与batch无关
        infinity: 无限迭代
        '''
        if isinstance(simgShape,int):
            simgShape = (simgShape,simgShape)
        self.handleImgGt = handleImgGt
        self.imggts = imggts
        self.simgShape = simgShape
        self.batch = batch
        self._iters = iters
        self.iters = self._iters
        self.infinity = infinity
        
        hh,ww = simgShape
        jpg,png = imggts[0]
        img = imread(jpg)
        h,w = img.shape[:2]
        if cache is None:
            cache = max(1,int(1e9/img.nbytes))
        cache = min(cache,len(imggts))
        self.maxPerCache = int(cache*(h*w)*1./(hh*ww))* timesPerRead/batch
        self.cache = cache
        self.n = len(imggts)
        self._times = max(1,int(round(self.n*1./cache/timesPerRead)))
        self.times = self._times
        self.totaln = self.sn = iters or int((h*w)*self.n*1./(hh*ww))
        self.willn = iters or self.maxPerCache*self.times*batch
        self.count = 0
        self.reset()
        
        self.bytes = img.nbytes
        argsStr = '''imggts=%s pics in dir: %s, 
        simgShape=%s, 
        handleImgGt=%s,
        batch=%s, cache=%s,iters=%s,
        timesPerRead=%s, infinity=%s'''%(self.n , os.path.dirname(jpg) or './', simgShape, handleImgGt,
                                 batch, cache,iters,
                                 timesPerRead,infinity)
        generatorStr = '''maxPerCache=%s, readTimes=%s
        Will generator maxPerCache*readTimes*batch=%s'''%(self.maxPerCache, self.times,
                                                          self.willn)
        if iters:
            generatorStr = 'Will generator iters=%s'%iters
        self.__describe = '''GenSimg(%s)
        
        Total imgs Could generator %s simgs,
        %s simgs.
        '''%(argsStr,self.totaln,
             generatorStr,)
    def reset(self):
        if (self.times<=0 and self.iters is None) and not self.infinity:
            self.times = self._times
            raise StopIteration
        self.now = self.maxPerCache
        inds = np.random.choice(list(range(len(self.imggts))),self.cache,replace=False)
        datas = {}
        for ind in inds:
            jpg,png = self.imggts[ind]
            img,gt = imread(jpg),imread(png)
            datas[jpg] = img,gt
        self.data = self.datas = datas
        self.times -= 1
    def __next__(self):
        self.count += 1
        if (self.iters is not None) and not self.infinity:
            if self.iters <= 0:
                self.iters = self._iters
                raise StopIteration
            self.iters -= self.batch
        if self.now <= 0:
            self.reset()
        self.now -= 1
        hh,ww = self.simgShape
        datas = self.datas
        imgs, gts = [], []
        for t in range(self.batch):
            img,gt = datas[np.random.choice(list(datas.keys()),1,replace=False)[0]]
            h,w = img.shape[:2]
            i= np.random.randint(h-hh+1)
            j= np.random.randint(w-ww+1)
            (img,gt) =  img[i:i+hh,j:j+ww],gt[i:i+hh,j:j+ww]
            imgs.append(img), gts.append(gt)
        (imgs,gts) = list(map(np.array,(imgs,gts)))
        if self.handleImgGt:
            return self.handleImgGt(imgs,gts)
        return (imgs,gts)
    @property
    def imgs(self):
        return [img for img,gt in list(self.datas.values())]
    @property
    def gts(self):
        return [gt for img,gt in list(self.datas.values())]
    def __str__(self):
        batch = self.batch
        n = len(self.datas)
        return self.__describe + \
        '''
    status:
        iter  in %s/%s(%.2f)
        batch in %s/%s(%.2f)
        cache imgs: %s
        cache size: %.2f MB
        '''%(self.count*batch,self.willn,self.count*1.*batch/self.willn,
            self.count,self._times*self.maxPerCache,
            self.count*1./(self._times*self.maxPerCache),
            n, (n*self.bytes/2**20))
    def __len__(self):
        return self.willn
    __repr__ = __str__

if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from ..ylsys import py2
from ..ylcompat import istype

import sys
import inspect
from collections import defaultdict

def listToBatch(listt, batch):
    '''
    将一段序列按照每batch个元素组成一组
    
    >>> listToBatch(range(8),3)
    [(0, 1, 2), (3, 4, 5), (6, 7)]
    '''
    n = len(listt)
    left = n % batch
    if left:
        ind  = n - left
        listt, tail = listt[:ind], tuple(listt[ind:])
    ziped = list(zip(*[iter(listt)]*batch))
    if left:
        ziped.append(tail)
    return ziped

class _SliceToInt():
    ''' convert float to int when in slice
    >>> range(5)[sliceInt[3.3]]
    3
    
    >>> range(5)[sliceInt[-0.1: 3.3]]
    range(0, 3)
    '''
    def intround(self, v):
        if v is None:
            return v
        return int(round(v))
    def __intSlice(self, s):
        intround = self.intround
        return slice(intround(s.start), intround(s.stop), intround(s.step),)
    def __getitem__(self, index):
        def f(t):
            if isinstance(t, tuple):
                return tuple(f(i) for i in t)
            else:
                if isinstance(t, slice):
                    return self.__intSlice(t)
                if isinstance(t, float):
                    return int(round(t))
                return t
        if isinstance(index, (tuple, slice)):
            return f(index)
        return int(round(index))
    @staticmethod
    def test():
        print(range(5)[sliceInt[3.3]])
        print(range(5)[sliceInt[3.3:]])
sliceInt = _SliceToInt()

class sliceLimit():
    ''' limit the value in slice by given numpy array
    
    >>> rr = np.zeros((4,4,4))
    >>> sliceLimit(rr)[-1:2, :1000]
    r[0:2, :4, :]
    
    >>> sliceLimit(rr, True)[-1:2, :1000]
    (slice(0, 2, None), slice(None, 4, None))
    
    >>> sliceLimit(rr, True)[-1:2, ..., :1000]
    (slice(0, 2, None), Ellipsis, slice(None, 4, None))
    '''
    def __init__(self, arr_or_shape, returnSlice=False):
        import numpy as np
        
        if not isinstance(arr_or_shape, np.ndarray) or \
        (arr_or_shape.ndim == 1 and arr_or_shape.size<=4):
            shape = arr_or_shape
            returnSlice = True
        else:
            self.arr = arr_or_shape
            shape = arr_or_shape.shape
        self.shape = shape
        self.returnSlice = returnSlice
    def getitem(self, index):
        shape = self.shape
        def sliceTupleLimit(slicee, maxx):
            if not isinstance(slicee, slice):
                return slicee
            return slice(
                    None if slicee.start is None else max(0, slicee.start), 
                    None if slicee.stop is None else min(maxx, slicee.stop), 
                    slicee.step)
        if isinstance(index, slice):
            return sliceTupleLimit(index, shape[0])
        if isinstance(index, tuple):
            splitInd = index.index(Ellipsis) if Ellipsis in index else len(index)
            left = tuple(sliceTupleLimit(slicee, n)  for slicee, n  in zip(index[:splitInd], shape))    
            right = tuple(sliceTupleLimit(slicee, n)  for slicee, n  in zip(index[splitInd+1:][::-1], shape[::-1]))[::-1]
            middle = (Ellipsis,) if Ellipsis in index else ()
            return left + middle + right
        return index
    def __getitem__(self, index):
        new_index = self.getitem(index)
        if self.returnSlice:
            return new_index
        return self.arr[new_index]
    @staticmethod
    def test():
        import numpy as np
        rr = np.zeros((4,4,4))
        print(sliceLimit(rr, True)[-1:2, :1000])
        print(sliceLimit(rr, True)[-3:80, ..., -3:80])

class Ll(list):
    '''
    a tool for convenient list operate
    
    1. quick exec `list(range(int))` buy ll*int:
    >>> ll*4
    [0, 1, 2, 3] 
    
    2. `ll-Iterable`, `ll/Iterable` is a convenient way to exec `list(Iterable)`
    '''
    def __init__(self):
        list.__init__(self, [0, 1])
    def __sub__(self, intOrLazyIter):
        if isinstance(intOrLazyIter, int):
            return self * intOrLazyIter
        return list(intOrLazyIter)
    def __mul__(self, intt):
        return list(range(intt))
    __call__ = __truediv__ = __div__ = __sub__
ll = Ll()

CALL_CLASS_CACHE={}
def addCall(instance):
    '''
    instance增加__call__ 返回自己
    '''
    t = type(instance)
    if t not in CALL_CLASS_CACHE:
        class T(t):
            def __call__(self):
                return self
        CALL_CLASS_CACHE[t]=T
    return CALL_CLASS_CACHE[t](instance)


fun_add_magic_paras = defaultdict(lambda :{})
class FunAddMagicMethodCore(object):
    '''magic 未解之谜 疑惑
    
    >>> z=FunAddMagicMethod(zip)
    >>> isinstance(z,type(zip)) => True
    >>> isinstance(z,object) => True
    '''
    def __init__(self, fun, returnArg=False):
        fun_add_magic_paras[id(self)]['fun'] = fun
        fun_add_magic_paras[id(self)]['returnArg'] = returnArg
    def __call__(self, *args, **kv):
        resoult = fun_add_magic_paras[id(self)]['fun']( *args, **kv)
        if fun_add_magic_paras[id(self)]['returnArg'] and len(args)==1:
            return args[0]
        return resoult
    __sub__ = __call__
    __mul__ = __call__
    __pow__ = __call__
#    __add__ = __call__
#    __eq__ = __call__
#    __invert__ = __call__
    def __div__(self, *args, **kv):
        fun_add_magic_paras[id(self)]['fun']( *args, **kv)
        if len(args)==1:
            return args[0]
        return args
    __truediv__ = __div__
    def __str__(self):
        return 'FunAddMagicMethod(%s)'%fun_add_magic_paras[id(self)]['fun']
    __repr__ = __str__
class FunAddMagicMethod(FunAddMagicMethodCore):
    '''
    add magic method to any callable instance or types,
    then you can call they buy sub, mul, pow operator to call they conveniently
    
    >>> from math import sqrt
    >>> magic_sqrt = mf - sqrt
    >>> (magic_sqrt-9, magic_sqrt*9, magic_sqrt**9) 
    (3.0, 3.0, 3.0)
    
    >>> magic_sqrt/9
    9
    
    the div operator is speacial, cause use div will return arg self
    
    magic_fun (- | * | **) arg => return fun(arg) 
    
    magic_fun / arg = exec fun(arg) but return arg itself
    
    
    
    将函数变为带有魔法函数 且可以__call__的对象
    fun : 需要增加为魔法对象的函数
    self - arg = fun(arg)
    self / arg = fun(arg) return arg
    '''
    def __getattribute__(self, name=None, *l):
        fun = fun_add_magic_paras[id(self)]['fun']
#        from boxx import gg, tree
        if name in dir(fun):
#            tree-[gg.n/name ,gg.f/fun, gg.l/l]
            return getattr(fun,name, *l)
        if name in ['__call__', '__class__', '__delattr__', '__dict__', '__dir__', '__div__', '__doc__', '__eq__', '__ge__', '__getattribute__', '__gt__', '__hash__', '__init__', '__init_subclass__', '__le__', '__lt__', '__module__', '__mul__', '__ne__', '__new__', '__pow__', '__reduce__', '__repr__', '__setattr__', '__str__', '__sub__', '__truediv__']:
            return FunAddMagicMethodCore.__getattribute__(self, name, *l)
        return getattr(fun,name, *l)
mf = FunAddMagicMethod(FunAddMagicMethod)

# invalid a function
fnone = mf(lambda *l, **kv: l[0] if len(l)==1 else l)

class dicto(dict):
    '''
    a subclass of dict for convenient, like object in JavaScript
    
    >>> d = dicto(a=0)
    >>> d.b = 1
    >>> print(d)
    {'a': 0, 'b': 1}
    
    类似 JavaScript 对象的超级字典，可以动态添加属性
    其中 dicto.keys,dicto.values,dicto.items 添加了property
    这个反射机制是试出来的 原理还不清楚
    '''
    def __init__(self, *l, **kv):
        dict.__init__(self,*l,**kv)
    def __getattribute__(self, name=None):
        if name in dir(dict):
            return dict.__getattribute__(self,name)
        if name in self:
            return self[name]
        return None
    def __setattr__(self, name, v):
        if name in dir(dict):
            return dict.__setattr__(self,name,v)
        self[name] = v
    if py2:
        @property
        def keys(self):
            return addCall(dict.keys(self))
        @property
        def values(self):
            return addCall(dict.values(self))
        @property
        def items(self):
            return addCall(dict.items(self))
SuperDict = dicto

def _yield():
    yield 0
generator = type(_yield())

def dicToObj(dic):  
    '''
    将 dict 转换为易于调用的 Object
    '''
    top = type('MyObject'.encode('utf-8'), (object,), dic)  
    seqs = tuple, list, set, frozenset  
    for i, j in list(dic.items()):  
        if isinstance(j, dict):  
            setattr(top, i, dicToObj(j))  
        elif isinstance(j, seqs):  
            setattr(top, i,   
                type(j)(dicToObj(sj) if isinstance(sj, dict) else sj for sj in j))  
        else:  
            setattr(top, i, j)  
    return top  
    

def typeNameOf(classOrType):
    '''
    以str 返回classOrType的所属类别  
    
    >>> typeNameOf(dict) 
    u'dict'
    '''
    ss = str(classOrType).split("'")
    if len(ss)>=3:
        return ss[-2]
    return str(classOrType)

def typestr(instance):
    '''
    以str 返回instance的所属类别  
    
    >>> typestr({}) 
    u'dict'
    '''
    return typeNameOf(type(instance))
typestr = FunAddMagicMethod(typestr)

def strMethodForDiraAttrs(self, pattern='^[^_]'):
    '''
    the default of __str__ method in Class
    
    will return string of dira(self) 
    
    '''
    from boxx import dira, PrintStrCollect
    printf = PrintStrCollect()
    dira(self, pattern=pattern, printf=printf, printClassFathers=False)
    s = str(printf)
    if py2:
        s = s[s.index(u'└'.encode('utf-8')):]
        return s
    s = s[s.index('└'):]
    return s

def nextiter(iterr, raiseException=True):
    '''
    do next(iter(iterAble)) then return resoult
    
    while iterr is empty and raiseException is False, just return  '【Iterable is empty!】'
    '''
    re = default = '【Iterable is empty!】'
    for i in iterr:
        re = i
        break
    if raiseException and re is default:
        raise StopIteration('Iterable is empty!')
    return re
nextiter = FunAddMagicMethod(nextiter)

def getfathers(objOrType):
    '''
    获得 objOrType 的所有父类，以 tuple 返回
    
    Parameters
    ----------
    objOrType : obj or type (includ classobj in python 2)
        if is obj, 则自动转换为 type(obj)
    '''
    if not istype(objOrType):
        objOrType = type(objOrType)
    return inspect.getmro((objOrType))
getfathers = FunAddMagicMethod(getfathers)

def getfather(objOrType): 
    '''
    获得 objOrType 的父类
    
    Parameters
    ----------
    objOrType : obj or type (includ classobj in python 2)
        if is obj, 则自动转换为 type(obj)
    '''
    return getfathers(objOrType)[0]
getfather = FunAddMagicMethod(getfather)
    

def setself(self=None):
    '''
    set all method(*args)  to self.__dict__ 
    
    >>> class A():
    ...     def __init__(self, attr='attr'):
    ...         setself()
    >>> a=A()
    >>> a.attr
    'attr'
    '''
    local = sys._getframe(1).f_locals
    if self is None:
        self = local['self']
    self.__dict__.update(local)
#    tree-local

def unfoldself(self=None):
    '''
    set all self.__dict__ to locals()
    
    >>> class A():
    ...     def __init__(self):
    ...         self.attr='attr'
    ...     def get_attr(self):
    ...         unfoldself()
    ...         return attr
    >>> a=A()
    >>> a.get_attr()
    'attr'
    
    ps. if vars name in locals(), the vars won't be cover
    '''
    local = sys._getframe(1).f_locals
    if self is None:
        self = local['self']
    for k, v in self.__dict__.items():
        if k not in local:
            local[k] = v

class withfun():
    '''
    Convenient way to use `with statement` without build a Class
    enterFun and exitFun are no parameter function
    
    Parameters
    ----------
    enterFun : function, default None
        No parameter function
    exitFun : function, default None
        No parameter function or parameter are (exc_type, exc_value, exc_traceback)
    exception : bool, default None
        Whether send (exc_type, exc_value, exc_traceback) to exitFun
    '''
    def __init__(self, enterFun=None, exitFun=None, exception=None):
        self.enterFun = enterFun
        self.exitFun = exitFun
        self.exception = exception
        if exception is None and self.exitFun:
            self.exception = len(inspect.getargspec(self.exitFun).args) == 3
    def __enter__(self):
        if self.enterFun:
            return self.enterFun()
    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.exitFun:
            if self.exception:
                self.exitFun(exc_type, exc_value, exc_traceback)
            else:
                self.exitFun()


class withattr():
    """
    set attr or item in `with statement`, after __exit__ the obj or dict will recovery like befor 
    
    Parameters
    ----------
    obj : obj or dict
        the thing that will change attr or item in with statement
    attrs_or_key : dict or str
        dict: the attrs or items that will change during with statement
        str: the key of the attr
    value: anything, default None
        when attrs_or_key is string(key), this arg as value
    
    Usage
    ----------
    
    >>> with withattr(dict(), 'attr', 'value') as d:
    ...     print(d['attr'])
    "value"
    >>> 'attr' in d
    False
    
    >>> with withattr(dict(), {'attr':'value'}) as d:
    ...     print(d['attr'])
    "value"
    >>> 'attr' in d
    False
        
    ps. `withattr` will detect whther the obj is dict, then choose setattr or setitem.
    """

    def __init__(self, obj, attrs_or_key, value=None):
        self.obj = obj

        if isinstance(attrs_or_key, str):
            attrs = {attrs_or_key: value}
        else:
            attrs = attrs_or_key

        d = obj

        sett = (
            (lambda d, k, v: d.__setitem__(k, v)) if isinstance(obj, dict) else setattr
        )
        get = (lambda d, k: d[k]) if isinstance(obj, dict) else getattr
        pop = (lambda d, k: d.pop(k)) if isinstance(obj, dict) else delattr
        has = (lambda d, k: k in d) if isinstance(obj, dict) else hasattr

        def enter_func():
            self.old = {}
            for k, v in attrs.items():
                if has(d, k):
                    self.old[k] = get(d, k)
                sett(d, k, v)
            return d

        def exit_func():
            for k in attrs.keys():
                if k in self.old:
                    sett(d, k, self.old[k])
                else:
                    pop(d, k)

        self.enter_func = enter_func
        self.exit_func = exit_func
    def __enter__(self):
        return self.enter_func()
    def __exit__(self, exc_type, exc_value, exc_traceback):
        return self.exit_func()
    
def isinstancestr(obj, typeStrList):
    '''
    same as isinstance but only use type name strs to avoid `import torch.tensor` .etc
    
    Parameters
    ----------
    obj : object
        anything
    typeStrList : str or list of strs
        like tuple of types for isinstance
    
    Return
    ----------
    if True return the type name str
    '''
    if isinstance(typeStrList, str):
        typeStrList = [typeStrList]
    types = getfathers(obj)
    typestrs = list(map(typeNameOf, types))
    for t in typeStrList:
        if t in typestrs:
            typestrs.index(t)
            return t
    return False

if __name__ == "__main__":

    pass


# -*- coding: utf-8 -*-

from __future__ import unicode_literals, print_function

from ..tool.toolStructObj import FunAddMagicMethod, typeNameOf, typestr, dicto
from ..tool.toolStructObj import generator, nextiter, getfathers, isinstancestr
from ..tool.toolLog import log, PrintStrCollect, colorFormat, clf, tounicode, LogLoopTime, prettyClassFathers
from ..tool.toolLog import tabstr, getDoc, shortStr, discrib, strnum
from ..tool.toolFunction import mapmp, pipe
from ..tool.toolSystem import tryImport
from ..ylsys import tmpYl, pyi, py2, sysi
from ..ylsci.ylnp import isNumpyType
from ..ylcompat import interactivePlot, beforImportPlt

import os
import glob
import math
import types
import numpy as np
import skimage as sk
from functools import reduce
from collections import defaultdict
from operator import add
#cv2 = tryImport('cv2')

#from skimage import io as sio
#from skimage.io import imread
#from skimage.io import imsave
#from skimage import data as sda

def imsave(fname, arr, plugin=None, **plugin_args):
    '''
    same usage of skimage.io.imsave, for lazy import skimage.io and matplotlib.pyplot
    '''
    beforImportPlt()
    from skimage.io import imsave
    return imsave(fname, arr, plugin, **plugin_args)

def imread(fname, as_grey=False, plugin=None, flatten=None, **plugin_args):
    '''
    same usage of skimage.io.imread, for lazy import skimage.io and matplotlib.pyplot
    '''
    beforImportPlt()
    from skimage.io import imread
    if flatten is None:
        return imread(fname, as_grey, plugin, **plugin_args)
    return imread(fname, as_grey, plugin, flatten, **plugin_args)

#class FakeSkimageData(types.ModuleType): # raise SystemError: nameless module when dir(sda)
class FakeSkimageData():
    __all__ = ['load', 'astronaut', 'camera', 'checkerboard', 'chelsea', 'clock', 'coffee', 'coins', 'horse', 'hubble_deep_field', 'immunohistochemistry', 'logo', 'moon', 'page', 'text', 'rocket', 'stereo_motorcycle']
    def __init__(self):
        pass
    def __getattr__(self, k, *l):
        beforImportPlt()
        from skimage import data as sda
        return getattr(sda, k)
    def __call__(self):
        beforImportPlt()
        from skimage import data as sda
        return sda.astronaut()
sda = FakeSkimageData()

# randomm((m, n), max) => m*n matrix
# randomm(n, max) => n*n matrix
randomm = lambda shape,maxx:(np.random.random(
shape if (isinstance(shape,tuple) or isinstance(shape,list)
)else (shape,shape))*maxx).astype(int)
r = randomm(4,4)

def normalizing(arr):
    if isinstance(arr, np.ndarray) and not isNumpyType(arr, 'float'):
        arr = arr.astype(float)
    minn = arr.min()
    maxx = arr.max()
    if maxx == minn:
        return arr/maxx
    return (arr-minn)/(maxx-minn)
normalizing = FunAddMagicMethod(normalizing)
norma = normalizing

def warpResize(img, hw, interpolation=None):
    '''
    resize by cv2.warpAffine to avoid cv2.resize()'s BUG(not align when resize bigger)
    '''
    import cv2
    from ..ylsci import Vector
    if interpolation is None:
        interpolation = cv2.INTER_NEAREST
    h, w = hw
    rhw = Vector(img.shape[:2])
    dhw = hw/rhw
    M = np.array([[dhw.w*1.,0,0],[0,dhw.h*1,0]])
    dst = cv2.warpAffine(img, M, (w, h), flags=interpolation,)
    return dst

def resize(img, arg2, interpolation=None):
    '''
    resize the np.ndarray or torch.Tensor
        
    Parameters
    ----------
    arg2: float, int, shape, ndarray, torch.Tensor
        the size or target ndarray 
        
    '''
    hw = arg2
    if isinstance(arg2, np.ndarray):
        if not(arg2.ndim == 1 and arg2.shape == (2,)):
            hw = arg2.shape[:2]
    elif typestr(arg2) in ['torch.Tensor']:
        hw = arg2.shape[-2:]
    elif isinstance(arg2, (float, int)):
        if arg2 == 1:
            return img
        hw = img.shape[:2] if isinstance(img, np.ndarray) else img.shape[-2:]
        if isinstance(arg2, (float)) or arg2 <= 5:
            hw = [int(round(size * arg2)) for size in hw]
        else:
            hw = [int(round(size * arg2 / max(hw))) for size in hw]
    if isinstance(img, np.ndarray):
        is_bool8 = img.dtype == np.bool8
        if is_bool8:
            img = np.uint8(img)
        dst = warpResize(img, hw, interpolation=interpolation)
        if is_bool8:
            dst = np.bool8(dst)
    elif typestr(img) in ['torch.Tensor']:
        interpolation =  interpolation or 'nearest'
        from torch import nn
        dst = nn.functional.interpolate(img, tuple(hw), mode=interpolation)
    return dst

def uint8(img):
    '''将0～1的float或bool值的图片转换为uint8格式'''
    return ((img)*255.999).astype(np.uint8)

greyToRgb = lambda grey:grey.repeat(3).reshape(grey.shape+(3,)) 

def histEqualize(img):
    from skimage.exposure import equalize_hist
    img = equalize_hist(img)
    minn = img.min()
    maxx = img.max()
    if maxx == minn:
        return img
    return (img-minn)/(maxx-minn)
histEqualize = FunAddMagicMethod(histEqualize)

boolToIndex = lambda boolMa1d:np.arange(len(boolMa1d))[npa(boolMa1d).squeeze()>0]
boolToIndex = FunAddMagicMethod(boolToIndex)

def tprgb(ndarray):
    '''
    transpose to RGB, 将3*h*w的ndarray化为 h*w*3的RGB图
    即shape为 (...,3 , h, w)的ndarray转化为(..., h, w, 3)
    '''
    shape = ndarray.shape
    ndim = ndarray.ndim
    if ndim >= 3 and shape[-3] == 3:
        axes = list(range(ndim))[:-3]+[ndim-2,ndim-1,ndim-3]
        return ndarray.transpose(*axes)
    return ndarray
tprgb = FunAddMagicMethod(tprgb)


def padding(img, urdl=1):
    '''
    padding image for up, right, down, left
    '''
    if not isinstance(urdl, list):
        urdl = [urdl] * 4
    from ..tool import intround
    u, r, d, l = map(intround, urdl)
    h, w = img.shape[:2]
    hh = h + u + d
    ww = w + r + l
    bimg = np.zeros((hh, ww,) + img.shape[2:], img.dtype)
    bimg[u:hh-d, l:ww-r] = img
    return bimg

def toPng(img):
    '''
    add alpha channel to be a png picture
    '''
    if img.shape[-1] == 3:
        return np.append(img, np.ones(img.shape[:-1]+(1,), img.dtype)* (255 if isNumpyType(img, 'int') else 1), 2)
    return img

def torgb(img):
    '''
    try to transfer a tensor to normalized RGB image
    
    normalizing img value to 0~1
    and transpose (..., 3, w, h) to (..., w, h, 3)
    
    '''
    img = npa(img)
    if img.min() < 0:
        img = norma(img)
    return tprgb(img)

def frombgr(img):
    img = torgb(img)
    if img.ndim >= 3 and img.shape[-1] == 3 :
        img = img[...,[2,1,0]]
    return img

torgb = FunAddMagicMethod(torgb)

def mapp(f, matrix, need_i_j=False):
    '''
    for each item of a 2-D matrix
    return a new matrix consist of f:f(it) or f(it, i, j)
    性能差 尽量用y, x = np.mgrid[:10,:10]
    '''
    m, n = matrix.shape[:2]
    listt = [[None]*n for i in range(m)]
    for i in range(m):
        for j in range(n):
            it = matrix[i][j]
            listt[i][j] = f(it,i,j) if need_i_j else f(it)
    return np.array(listt)

def __torchVar2Tensor(t):
    '''
    同时兼容 PyTorch 3.0 和 4.0 的 Tensor 和 Var
    '''
    try:
        t = t.data
    except:
        pass
    finally:
        return t
__torchToNumpy = lambda x:__torchVar2Tensor(x).numpy()
__torchCudaToNumpy = lambda x:__torchVar2Tensor(x).cpu().numpy()
__todense = lambda x:x.todense()

typesToNumpyFuns = {
    'scipy.sparse.csr.csr_matrix':__todense,
    'scipy.sparse.coo.coo_matrix':__todense,
    'scipy.sparse.csc.csc_matrix':__todense,
    'PIL.Image.Image':lambda x:np.array(x),
    'mxnet.ndarray.NDArray':lambda x:x.asnumpy(),
    'mxnet.ndarray.ndarray.NDArray':lambda x:x.asnumpy(),
    
    'torch.Tensor':__torchCudaToNumpy,
    
    'torch.FloatTensor':__torchToNumpy,
    'torch.DoubleTensor':__torchToNumpy,
    'torch.IntTensor':__torchToNumpy,
    'torch.LongTensor':__torchToNumpy,
    'torch.ShortTensor':__torchToNumpy,
    'torch.ByteTensor':__torchToNumpy,
    'torch.HalfTensor':__torchToNumpy,
    'torch.CharTensor':__torchToNumpy,
    
    "torch.cuda.LongTensor":__torchCudaToNumpy,
    "torch.cuda.DoubleTensor":__torchCudaToNumpy,
    "torch.cuda.IntTensor":__torchCudaToNumpy,
    "torch.cuda.ShortTensor":__torchCudaToNumpy,
    "torch.cuda.ByteTensor":__torchCudaToNumpy,
    "torch.cuda.HalfTensor":__torchCudaToNumpy,
    "torch.cuda.CharTensor":__torchCudaToNumpy,
    "torch.cuda.FloatTensor":__torchCudaToNumpy,
    
    "torch.autograd.variable.Variable":lambda x:__torchCudaToNumpy(x.data),
    "torch.nn.parameter.Parameter":lambda x:__torchCudaToNumpy(x.data),
}

__strToNumpy = lambda x: imread(x) if os.path.isfile(x) else np.array(list(x))
__generatorToNumpy = lambda x:np.array(list(x))
tryToNumpyFunsForNpa = {
    "dict_values":__generatorToNumpy,
    "dict_keys":__generatorToNumpy,
    "dict_items":__generatorToNumpy,
    "map":__generatorToNumpy,
    "filter":__generatorToNumpy,
    "zip":__generatorToNumpy,
    
    "str":__strToNumpy,
    "unicode":__strToNumpy,
    }
tryToNumpyFunsForNpa.update(typesToNumpyFuns)

def npa(array):
    '''
    try to transfer other data to np.ndarray
    
    support types inculde:
        numpy, torch.tensor, mxnet.ndarray, PIL.Image
        list, tuple, dict, range, zip, map
        str, image_path
    
    Parameters
    ----------
    array : list/tuple, torch.Tensor, mxnet.NDArray ...
        support types in boxx.ylimg.ylimgTool.typesToNumpyFuns and boxx.ylimg.ylimgTool.tryToNumpyFunsForNpa
    '''
    if isinstance(array, np.ndarray):
        if type(array) is not np.ndarray:
            return np.array(array)
        return array
    
    typeNameForNpa = isinstancestr(array, tryToNumpyFunsForNpa)
    if typeNameForNpa:
        ndarray = tryToNumpyFunsForNpa[typeNameForNpa](array)
    else:
        ndarray = np.array(array)
    return ndarray
npa = FunAddMagicMethod(npa)


def discribArray(array):
    '''
    return str of discrib for array, include shape type max min
    '''
    typeName = typestr(array)
    array = npa(array)
    strr = (clf.r%tounicode(array.shape),
          clf.r%typeNameOf(array.dtype.type)[6:], 
          clf.r%typeName,
          clf.r%( strnum(array.max()) if (array.size) else 'Empty'), 
          clf.r%( strnum(array.min()) if (array.size) else 'Empty'), 
          clf.r%( strnum(array.mean()) if (array.size) else 'Empty') )
    return (('shape:%s type:(%s of %s) max: %s, min: %s, mean: %s'%tuple(strr)))
    
    
def prettyArray(array):
    '''
    return str of pretty discrib for array, include nan inf (shape type max min)  
    '''
    discrib = discribArray(array)
    array = npa(array)
    unique = np.unique(array)
    finiteInd = np.isfinite(array)
    nan = np.isnan(array).sum()
    
    discribNan = ''
    if not (finiteInd).all():
        finite = array[finiteInd]
        size = array.size
        nan = np.isnan(array).sum()
        nans = clf.p%'"nan":%s (%.2f%%), '%(nan,100.*nan/size) if nan else ''
        inf = np.isinf(array).sum()
        infs = clf.p%'"inf":%s (%.2f%%), '%(inf,100.*inf/size) if inf else ''
        discribNan = (clf.r%'\nNotice: ' + '%s%s finite max: %s, finite min: %s, finite mean: %s'%(
                nans,infs, len(finite) and strnum(finite.max()), len(finite) and strnum(finite.min()), 
                len(finite) and strnum(finite.mean())))
    if len(unique)<10:
        dic = defaultdict(lambda : 0)
        for i in array.ravel():
            dic[i] += 1
        listt = list(dic.items())
        listt.sort(key=lambda x:x[0])
        x = np.array([k for k,v in listt]).astype(float)
        if len(x) == 0:
            discrib += (clf.p%'\nEmpty array')
        elif len(x) == 1:
            discrib += (clf.p%'\nAll value is %s' % x[0])
        else:
            discrib += ('\nOnly %s unique values are %s'%(len(x), ', '.join([clf.p%v for v in x])))
    discrib += discribNan
    return discrib

#interactivePlot = lambda x:x
@interactivePlot
def plot(array, sort=False, maxline=10):
    '''
    plot line or top @maxline lines
    '''    
    import matplotlib.pyplot as plt
    plt.figure()
    if callable(array) and '__iter__' not in dir(array):
        x = np.linspace(np.e*-1.5,np.e*1.5,100)
        array = array(x)
    discrib = prettyArray(array).replace('\n','\n\n')
    print(discrib)
    array = npa(array).squeeze()
    n = array.shape[-1]
    if array.ndim >= 2:
        array = np.resize(array, (array.size//n, n))
        arrays = array[:maxline]
    else:
        arrays = [array]
    for arr in arrays:
        if sort:
            arr = sorted(arr, reverse=True)
        plt.plot(arr)
    plt.show()        
plot = FunAddMagicMethod(plot)

@interactivePlot
def loga(array):
    '''
    Analysis any array like thing .
    the shape, max, min, distribute of the array
    
    support numpy, torch.tensor, mxnet.ndarray, PIL.Image .etc
    '''
    import matplotlib.pyplot as plt
    plt.figure()
    discrib = prettyArray(array)
    print(discrib.replace('\n','\n\n'))
    
    array = npa(array)
    unique = np.unique(array)
    finiteInd = np.isfinite(array)
    if not (finiteInd).all():
        finite = array[finiteInd]
        data, x = np.histogram(finite,8)
    if len(unique)<10:
        dic = defaultdict(lambda : 0)
        for i in array.ravel():
            dic[i] += 1
        listt = list(dic.items())
        listt.sort(key=lambda x:x[0])
        data,x=[v for k,v in listt],np.array([k for k,v in listt]).astype(float)
        if len(x) <= 1:
            return
        width = (x[0]-x[1])*0.7
        x -=  (x[0]-x[1])*0.35
    elif not (finiteInd).all():
        x=x[1:]
        width = (x[0]-x[1])
    else:
        data, x = np.histogram(array.ravel(),8)
        x=x[1:]
        width = (x[0]-x[1])
    plt.plot(x, data, color = 'orange')
    plt.bar(x, data,width = width, alpha = 0.5, color = 'b')
    plt.show()
    
loga = FunAddMagicMethod(loga)

def ndarrayToImgLists(arr):
    '''
    将所有ndarray转换为imgList
    '''
    arr = np.squeeze(arr)
    ndim = arr.ndim
    if ndim <= 1:
        return []
    if arr.ndim==2 or (arr.ndim ==3 and arr.shape[-1] in [3,4]):
         return [arr]
    if arr.shape[-1] == 2 and arr.ndim >= 3: # 二分类情况下自动转换
        arr = arr.transpose(list(range(ndim))[:-3]+[ndim-1,ndim-3,ndim-2])
    imgdim = 3 if arr.shape[-1] in [3,4] else 2
    ls = list(arr)
    while ndim-1>imgdim:
        ls = reduce(add,list(map(list,ls)),[])
        ndim -=1
    return ls

def listToImgLists(l, res=None,doNumpy=ndarrayToImgLists):
    '''
    将 ndarray和list的混合结果树转换为 一维 img list
    '''
    if res is None:
        res = []
    for x in l:
        typeName = isinstancestr(x, typesToNumpyFuns)
        fathersStr = str(getfathers(x))
        if typeName:
            ndarray = typesToNumpyFuns[typeName](x)
            res.extend(doNumpy(ndarray))
        elif isinstance(x,(list,tuple)):
            listToImgLists(x,res=res,doNumpy=doNumpy)
        elif isinstance(x,dict):
            listToImgLists(list(x.values()),res=res,doNumpy=doNumpy)
        elif isinstance(x,np.ndarray):
            res.extend(doNumpy(x))
        elif ('torch.utils.data') in fathersStr or ('torchvision.datasets') in fathersStr:
            seq = unfoldTorchData(x, fathersStr)
            if seq is not False:
                listToImgLists(seq,res=res,doNumpy=doNumpy)
    return res

def _split_and_show(imgs, titles, split: int, **kv):
    showImgLists(imgs[:split], titles=titles[:split], **kv)
    showImgLists(imgs[split:], titles=titles[split:], **kv)

@interactivePlot
def showImgLists(imgs,**kv):
    import matplotlib.pyplot as plt
    n = len(imgs)
    titles = kv.pop("titles",[])

    if "ncols" in kv:
        ncols = kv["ncols"]
        if n >= 2*ncols:
            _split_and_show(imgs, titles, split=ncols, **kv)
            return
        elif n > ncols:
            _split_and_show(imgs, titles, split=math.ceil(n/2.), **kv)
            return
        del kv["ncols"]
    elif n == 4:
        _split_and_show(imgs, titles, split=2, **kv)
        return
    elif n > 4:
        _split_and_show(imgs, titles, split=3, **kv)
        return
    if "figsize" in kv:
        figsize = kv.pop("figsize")
        fig, axes = plt.subplots(ncols=n, figsize=figsize)
    else:
        fig, axes = plt.subplots(ncols=n)
    count = 0
    axes = [axes] if n==1 else axes
    for img in imgs:
        axes[count].imshow(img,**kv)
        axes[count].set_title(titles[count] if count < len(titles) else "")
        count += 1
    plt.show()
    
def show(*imgsAndFuns,**kv):
    '''
    show could find every image in complex struct and show they
    could sample images from torch.DataLoader and DataSet
    
    if imgsAndFuns inculde function. those functions will process all numpys befor imshow
    
    Parameters
    ----------
    imgsAndFuns : numpy/list/tuple/dict/torch.tensor/PIL.Image/function
        if imgsAndFuns inculde function . 
        those functions will process all numpys befor imshow
    kv : args
        args for plt.imshow
    找出一个复杂结构中的所有numpy 转换为对应的图片并plt.show()出来
    '''
    if 'cmap' not in kv:
        kv['cmap'] = 'gray'
    funs = [arg for arg in imgsAndFuns[1:] if callable(arg)]
    doNumpy = pipe(funs+[ndarrayToImgLists])
    imgls = listToImgLists(imgsAndFuns,doNumpy=doNumpy)
    imgls = [img for img in imgls if img.ndim >= 2 and min(img.shape) > 2]
    assert len(imgls)!=0,"function `show`'s args `imgs`  has no any np.ndarray that ndim >= 2! "
    showImgLists(imgls,**kv)
show = FunAddMagicMethod(show)


def showb(*arr, png=False):
    '''
    use shotwell to show picture
    Parameters
    ----------
    arr : np.ndarray or path
    '''
    
    if len(arr)!=1:
        list(map(lambda ia:showb(ia[1],tag=ia[0]),enumerate(arr)))
        return 
    arr = arr[0]
    if isinstance(arr,np.ndarray):
        if arr.ndim == 3 and arr.shape[-1] == 3 and not png:
            path = tmpYl + 'tmp-%s.jpg'%len(glob.glob(tmpYl + 'tmp-*.jpg'))
        else:
            path = tmpYl + 'tmp-%s.png'%len(glob.glob(tmpYl + 'tmp-*.png'))
        if arr.dtype == np.bool8:
            arr = np.uint8(arr) * 255
        imsave(path,arr)
        arr = path
    cmd = 'shotwell "%s" &'%arr
    if sysi.win:
        cmd = '"%s"'%arr
    os.system(cmd)
showb = FunAddMagicMethod(showb)


imgExtNames = ['jpg', 'jpeg', 'png', 'gif', 'tif', 'bmp']
def isImgFileName(fname):
    return '.' in fname and fname.split('.')[-1].lower() in imgExtNames

def shows(*imgs, png=False, coln=None):
    '''图片展示分析工具  使用浏览器同步显示同一图像的不同数据表示 如不同通道的image,gt,resoult 
    支持放大缩小与拖拽
    
    Parameters
    ----------
    imgs : list include path or np.ndarray 
        图片地址或图片np.ndarray 组成的list 要求所有图片宽高相同
    '''
    
    def _listToImgLists(l, res=None,doNumpy=ndarrayToImgLists):
        '''
        将 ndarray和list的混合结果树转换为 一维 img list
        '''
        if res is None:
            res = []
        for x in l:
            typeName = isinstancestr(x, typesToNumpyFuns)
            if typeName:
                ndarray = typesToNumpyFuns[typeName](x)
                res.extend(doNumpy(ndarray))
            elif isinstance(x,(list,tuple)):
                _listToImgLists(x,res=res,doNumpy=doNumpy)
            elif isinstance(x,dict):
                _listToImgLists(list(x.values()),res=res,doNumpy=doNumpy)
            elif isinstance(x,np.ndarray):
                res.extend(doNumpy(x))
            elif isinstance(x,str):
                if os.path.isfile(x) and isImgFileName(x):
                    res.append(x)
        return res
    
    
    funs = [arg for arg in imgs[1:] if callable(arg)]
    doNumpy = pipe(funs+[ndarrayToImgLists])
    imgs = _listToImgLists(imgs,doNumpy=doNumpy)
    biggest_img = max(imgs, key=lambda x: x.shape[0]*x.shape[1] if isinstance(x, np.ndarray) else 0)
    biggest_hw = biggest_img.shape[:2]
    
    
    showsDir = os.path.join(tmpYl, 'shows')
    dirr = showsDir + '/shows-%03d.html' % len(glob.glob(showsDir + '/shows-*.html'))
    os.makedirs(dirr, exist_ok=True)
    
    paths = []
    for idx, x in enumerate(imgs):
        if isinstance(x,str):
            x = imread(x)
        if x.ndim == 3 and x.shape[-1] == 3 and not png:
            fname = '%s.jpg'%idx
        else:
            fname = '%s.png'%idx
        imgp = os.path.join(dirr, fname)
        if x.dtype == np.bool8:
            x = x * 255
        if x.shape[:2] != biggest_hw:
            x = resize(x, biggest_hw)
        imsave(imgp, x)
        paths.append(fname)
    htmlp = os.path.join(dirr, 'index.html')
    from .showImgsInBrowser import showImgsInBrowser
    showImgsInBrowser(paths, htmlp, coln=coln)
shows = FunAddMagicMethod(shows)


    
def __torchShape(x):
    if x.shape:
        s = "%s of %s @ %s"%(tuple(x.shape), x.type(), str(x.device))
    else:
        s = "%s of %s @ %s"%(strnum(float(x)), x.type(), str(x.device))
    return colorFormat.r%s

__numpy_struch_log_fun = lambda x:colorFormat.r%('%s%s'%
                                    (str(x.shape).replace('L,','').replace('L',''),x.dtype))

StructLogFuns = {
    'list':lambda x:colorFormat.b%('list  %d'%len(x)),
    'tuple':lambda x:colorFormat.b%('tuple %d'%len(x)),
    'dict':lambda x:colorFormat.b%('dict  %s'%len(x)),
    'mappingproxy':lambda x:colorFormat.b%('mappingproxy  %s'%len(x)),
    'set':lambda x:(colorFormat.r%'set %s = '%len(x) + colorFormat.b%str(x)),
    'collections.defaultdict':lambda x:colorFormat.b%('defaultDict  %s'%len(x)),
    'dicto':lambda x:colorFormat.b%('dicto  %s'%len(x)),
    'tool.toolStructObj.dicto':lambda x:colorFormat.b%('dicto  %s'%len(x)),
    'tool.toolLog.SuperG':lambda x:colorFormat.b%('SuperG  %s'%len(x)),
    'collections.OrderedDict':lambda x:colorFormat.b%('OrderedDict  %s'%len(x)),
    'dictoSub':lambda x:colorFormat.b%('dictoSub  %s'%len(x)),
    
    'numpy.ndarray':__numpy_struch_log_fun,
    'imageio.core.util.Array':__numpy_struch_log_fun,
    'scipy.sparse.csr.csr_matrix':lambda x:__numpy_struch_log_fun(x) + colorFormat.r%" of sparse.csr",
    'scipy.sparse.coo.coo_matrix':lambda x:__numpy_struch_log_fun(x) + colorFormat.r%" of sparse.coo",
    'scipy.sparse.csc.csc_matrix':lambda x:__numpy_struch_log_fun(x) + colorFormat.r%" of sparse.csc",
    
    'torch.Tensor':__torchShape,
    
    'torch.FloatTensor':__torchShape,
    'torch.DoubleTensor':__torchShape,
    'torch.IntTensor':__torchShape,
    'torch.LongTensor':__torchShape,
    'torch.ShortTensor':__torchShape,
    'torch.ByteTensor':__torchShape,
    'torch.HalfTensor':__torchShape,
    'torch.CharTensor':__torchShape,
    
    "torch.cuda.LongTensor":__torchShape,
    "torch.cuda.DoubleTensor":__torchShape,
    "torch.cuda.IntTensor":__torchShape,
    "torch.cuda.ShortTensor":__torchShape,
    "torch.cuda.ByteTensor":__torchShape,
    "torch.cuda.HalfTensor":__torchShape,
    "torch.cuda.CharTensor":__torchShape,
    "torch.cuda.FloatTensor":__torchShape,
    "torch.autograd.variable.Variable":lambda x:__torchShape(x.data),
    "torch.nn.parameter.Parameter":lambda x:__torchShape(x.data),
    
    'torch.utils.data.dataloader.DataLoader':lambda x:(colorFormat.b%'DataLoader(len=%d, batch=%d, worker=%d)'%
                                                       (len(x.dataset), x.batch_size, x.num_workers)),
    
    'mxnet.ndarray.NDArray':lambda x:colorFormat.r % '%s of mxnet.%s'%(str(x.shape), str(x.dtype)),
    'mxnet.ndarray.ndarray.NDArray':lambda x:colorFormat.r % '%s of mxnet.%s'%(str(x.shape), str(x.dtype)),
    
    'pandas.core.frame.DataFrame':lambda x:clf.r%'DataFrame(col=%d, Index=[%s], dtype="%s")'%(x.shape[0], ', '.join(map(str,x.columns)), x.columns.dtype)
    }

def discribOfInstance(instance,leafColor=None,MAX_LEN=45):
    typee = type(instance)
    typen = typeNameOf(typee)
#    print typen,typen in StructLogFuns
    if isinstance(instance,dicto) and typen not in StructLogFuns:
        typen = 'dictoSub'
    if typen in StructLogFuns:
        s = StructLogFuns[typen](instance)
        return shortStr(s, MAX_LEN+18)
    s = tounicode(instance)
    if len(s) > MAX_LEN:
        s = s[:MAX_LEN-3]+'...'
    return (leafColor or '%s')%s


def unfoldTorchData(seq, fathersStr=''):
    '''
    unfold torch.Dataset and Dataloader use next(iter()) for tree  
    '''
    import torch
    if isinstance(seq, torch.utils.data.DataLoader) or 'DataLoader' in fathersStr:
        seq = [('DataLoader.next', nextiter(seq, raiseException=False))]
    elif isinstance(seq, torch.utils.data.Dataset) or 'Dataset' in fathersStr:
        seq = [(colorFormat.b%'Dataset[0/%d]'%len(seq), seq[0])]
    else:
        return False
    return seq

MappingProxyType = dict if py2 else types.MappingProxyType 
    
iterAbleTypes = (list,tuple,dict,types.GeneratorType, MappingProxyType)
def unfoldAble(seq):
    '''
    能展开的 object 
    '''
    if isinstance(seq,iterAbleTypes) :
        if isinstance(seq,(list,tuple)):
            seq = list(enumerate(seq))
        elif isinstance(seq,(dict, MappingProxyType)):
            seq = list(seq.items())
        elif isinstance(seq, types.GeneratorType):
            seq = [('Generator.next', nextiter(seq, raiseException=False))]
        return seq
    fathersStr = str(getfathers(seq))
    if ('torch.utils.data') in fathersStr or ('torchvision.datasets') in fathersStr:
        return unfoldTorchData(seq, fathersStr)
    return False

class HiddenForTree():
    def __init__(self, lenn, maxprint):
        self.n = lenn
        self.m = maxprint
        self.s = colorFormat.r%'Hidden %s of all %d'%((lenn-maxprint//2*2), lenn)
        self.repeat = max(4,min(1,maxprint//2))
    def __str__(self):
        return self.s
    def strr(self, leftStr):
        lineTempl = ''.join(leftStr)
        lineTempl += '├── '
        half = (lineTempl + colorFormat.r%'···\n')*self.repeat
        mid = lineTempl + colorFormat.b%self.s+'\n'
        s = half + mid + half
        return s 
    
def tree(seq,maxprint=50,deep=None,logLen=45,printf=log,leafColor='\x1b[31m%s\x1b[0m',__key='/',__leftStr=None, __islast=None,__deepNow=0, __sets=None):
    '''
    类似bash中的tree命令 
    直观地查看list, tuple, dict, numpy, tensor, dataset, dataloader 等组成的树的每一层结构   
    可迭代部分用蓝色 叶子用红色打印   
    以下命令可以查看支持的数据结构类型   
    
    >>> tree(boxx.ylimg.ylimgTool.StructLogFuns)
    
    Parameters
    ----------
    seq : list or tuple or dict or numpy or tensor or torch.Dataset or torch.Dataloader or any Object
        打印出 以树结构展开所有可迭代部分
    maxprint : int, default 50
        每个 seq 内部，最大允许的数目 默认最多展示 50 个   
        若 bool(maxprint) 为 False 则不做限制
    deep : int, default None
        能显示的最深深度, 默认不限制
    logLen : int, default 45
        能显示的最长字符数
    printf : function, default print function
        a function that could replace print 
    
    ps.可在StructLogFuns中 新增类别
    '''
    if deep and __deepNow > deep:
        return
    if __leftStr is None:
        __leftStr = [] 
        __islast = 1
        __sets = set()
    if maxprint and isinstance(seq, HiddenForTree):
        printf(seq.strr(__leftStr), end='')
        return 
#    s = StructLogFuns.get(type(seq),lambda x:colorFormat.r%tounicode(x)[:60])(seq)
    try:
        s = discribOfInstance(seq,leafColor=leafColor,MAX_LEN=logLen)
    except Exception as e:
        s = colorFormat.r%"【%s】"%e.__repr__()
    s = s.replace('\n','↳')
#    printf ''.join(__leftStr)+u'├── '+tounicode(k)+': '+s
    try:  # Support tree({t:r})
        str_key = discribOfInstance(__key, MAX_LEN=logLen)
    except Exception as e:
        str_key = colorFormat.r%"【%s】"%e.__repr__()
    str_key = str_key.replace('\n','↳')
        
    printf('%s%s %s: %s'%(''.join(__leftStr), '└──' if __islast else '├──',tounicode(str_key),s))
    
    unfold = unfoldAble(seq)
    if unfold is False :
        return 
    else:
        if id(seq) in __sets:
            seq=[(colorFormat.p%'【printed befor】','')]
        else:
            __sets.add(id(seq))
            seq = unfold
    __leftStr.append('    'if __islast else '│   ')
    if maxprint : 
        lenn = len(seq)
        if lenn > maxprint:
            head = maxprint//2
            seq = seq[:head] + [('HiddenForTree',HiddenForTree(lenn=lenn, maxprint=maxprint))] + seq[-head:]
    for i,kv in enumerate(seq):
        __key,v = kv
        tree(v,maxprint=maxprint,deep=deep,logLen=logLen, printf=printf,leafColor=leafColor, __key=__key, __leftStr=__leftStr, 
             __islast=(i==len(seq)-1),__deepNow=__deepNow+1,__sets=__sets)
    __leftStr.pop()
tree = FunAddMagicMethod(tree)

def __typee__(x):
    return tounicode(type(x)).split("'")[1]

logModFuns = {
 type(os):lambda x:colorFormat.r%(__typee__(x)),
 }
logMod = lambda mod:logModFuns.get(type(mod),lambda x:colorFormat.b%tounicode(__typee__(x))[:60])(mod)
def treem(mod, types=None, deep=None, __leftStrs=None, __name='/', 
          __islast=None, __deepNow=0, __rootDir=None, __sets=None):
    '''
    类似bash中的tree命令 查看module及子module的每一层结构
    一目了然module的结构和api
    module用红色显示 其余部分用蓝色显示
    >>>treem(os)
    
    Parameters
    ----------
    mod : module
        显示出 以树结构展开module及其子module
        type(mod) should => module
    types : list of types, default None
        需要显示出来的类型(module 类型会自动添加)
        默认显示所有类型
        ps.为了使用deep更简便 若`treem(mod, [int])` 
           则自动转换为`treem(mod, types=None, deep=[int])`
    deep : int, default None
        能显示的最深深度
        默认不限制
    '''
    if isinstance(types, int):
        deep = types
        types = None
    if deep and __deepNow > deep:
        return
    if __leftStrs is None:
        __leftStrs = [] 
        __islast = 1
        if '__file__' not in dir(mod): 
            print('type(%s: %s) is not module!'%(logMod(mod),tounicode(mod)))
            return 
        __rootDir = os.path.dirname(mod.__file__)
        __sets = set()
    typeStr = logMod(mod)
    modKinds = ['','','(not sub-module)','(printed befor)']
    modKind = 0
    if isinstance(mod,type(os)):
        if '__name__' in dir(mod) :
            __name = mod.__name__
        modKind = 1
        dirMod = dir(mod)
        if  mod in __sets:
            modKind = 3 
        elif '__file__' not in dir(mod) or __rootDir not in mod.__file__:
            modKind = 2
    names = (tounicode(__name)+('   ' if modKind<2 else '  ·' )*20)[:40]+modKinds[modKind]
    
    print('%s%s %s: %s'%(''.join(__leftStrs), '└──' if __islast else '├──',typeStr,names))
    
    if modKind !=1:
        return
    __sets.add(mod)
    dirMod = [i for i in dirMod if i not in ['__name__','__file__','unicode_literals']]
    if types is not None:
        dirMod=[i for i in dirMod if type(mod.__getattribute__(i)) in  list(types)+[type(os)]] 
    __leftStrs.append('    'if __islast else '│   ')
    for i,name in enumerate(dirMod):
        e = mod.__getattribute__(name)
        treem(e,types,deep,__leftStrs,name,__islast=(i==len(dirMod)-1),__deepNow=__deepNow+1,__rootDir=__rootDir,__sets=__sets)
    __leftStrs.pop()
treem = FunAddMagicMethod(treem)

def getFunDocForDira(f):
    d = getDoc(f)
    if d:
        return ' : %s'%(colorFormat.black%tounicode(d.strip()))
    return ''

attrLogFuns = {
'method-wrapper': lambda x:'method-wrapper%s'%getFunDocForDira(x),
'builtin_function_or_method':lambda x:'builtin-method%s'%getFunDocForDira(x),
'instancemethod':lambda x:'instancemethod%s'%getFunDocForDira(x),
'buffer':lambda x:'buffer : %s'%(colorFormat.b%tounicode(x)),
}

def __dira(seq,instance=None, maxDocLen=50, deep=None, printf=print, __leftStr=None,__key='/',__islast=None,__deepNow=0, __sets=None):
    '''
    类似bash中的tree命令 简单查看instance的 __attrs__ 组成的树的结构
    attr name用红色；str(instance.attr)用蓝色；
    如果attr 为instancemethod，builtin_function_or_method，method-wrapper之一
    instance.attr.__doc__用黑色 
    
    ps.可在__attrLogFuns中 新增常见类别
    '''
    if deep and __deepNow > deep:
        return
    s = discribOfInstance(seq,colorFormat.b,MAX_LEN=maxDocLen)
    if maxDocLen < 100:
        s = s.replace('\n','↳')
    if __leftStr is None:
        __leftStr = [] 
        __islast = 1
        __sets = set()
        doc = '' if instance is None else getFunDocForDira(instance)
        if len(doc) > maxDocLen:
            doc = doc[:maxDocLen-3]+'...'
        s=colorFormat.b%('%d attrs%s'%(len(seq), doc.replace('\n','↳').replace(' :',',',1)))
#    printf ''.join(__leftStr)+u'├── '+tounicode(k)+': '+s
    printf('%s%s %s: %s'%(''.join(__leftStr), '└──' if __islast else '├──',colorFormat.r%tounicode(__key),s))
    if isinstance(seq,(list,tuple,dict)) :
        if id(seq) in __sets:
            seq=[(colorFormat.p%'【printed befor】','')]
        else:
            __sets.add(id(seq))
            if isinstance(seq,(list,tuple)):
                seq = list(enumerate(seq))
            elif isinstance(seq,(dict)):
                seq = list(seq.items())
                try: # key may not be sort
                    seq.sort(key=lambda x:x[0])
                except:
                    pass
    else:
        return 
    __leftStr.append('    'if __islast else '│   ')
    seq =  [(k,filterMethodName(k,v)) for k,v in seq]
    for i,kv in enumerate(seq):
        __key,v = kv
        __dira(v,maxDocLen=maxDocLen,deep=deep, printf=printf,__leftStr=__leftStr,__key=__key,
               __islast=(i==len(seq)-1), __deepNow=__deepNow+1,__sets=__sets)#leafColor=colorFormat.black)
    __leftStr.pop()

def filterMethodName(attrName, attr):
    typee = type(attr)
    typn = typeNameOf(typee)
    if typn in attrLogFuns:
        return attrLogFuns[typn](attr)
    if attrName in ('__globals__', 'func_globals'):
        return colorFormat.b%('【globals-dict %d omitted】'%len(attr))
    elif attrName in ('__builtins__', ) and isinstance(attr, dict):
        return colorFormat.b%('【builtins-dict %d omitted】'%len(attr))
    elif attrName in ('__all__',):
        return colorFormat.b%('【all-list %d omitted】'%len(attr))
    elif attrName == ('f_builtins'):
        return colorFormat.b%('【f_builtins %d omitted】'%len(attr))
    return attr

def dira(instance, pattern=None, deep=None, maxDocLen=50, printf=print, printClassFathers=True):
    '''
    `dira(x)` is supplement of `dir(x)`. 
    `dira(x)` will pretty print `x`'s all attribute in tree struct.    
    And `dira(x)` will print `x`'s Father Classes too.    
    
    Parameters
    ----------
    instance : Anything
        Anything, Instance better
    pattern : re.pattern
        use re.search to filter attrs
    deep : int, default None
        max deep of struct object
    maxDocLen : int, default 50
        max len of doc
    printf : function, default print function
        a function that could replace print 
        
    P.S.unfold ('__globals__', 'func_globals', __builtins__, __all__, f_builtins)
    
    
    Old Chinese
    ---------
    
    以树的结构 分析instance的所有 attrs, 并展示父类的继承链
    attr name用红色；str(instance.attr)用蓝色；
    如果attr 为instancemethod，builtin_function_or_method，method-wrapper之一
    instance.attr.__doc__用黑色 
    
    Parameters
    ----------
    instance : Anything
        Anything, Instance better
    pattern : re.pattern
        用于匹配re.search参数 进行filter
        ps.为了使用deep更简便 若`dira(instance, [int])` 
           则自动转换为`dira(instance, pattern=None, deep=[int])`
    maxDocLen : int, default 50
        若有文档显示文档的字数长度
    deep : int, default None
        能显示的最深深度, 默认不限制
    printf : function, default print function
        a function that could replace print 
        
    ps.可在__attrLogFuns中 新增常见类别
    pps.不展开 ('__globals__', 'func_globals', __builtins__, __all__, f_builtins)
    '''
    if printClassFathers:
        s = prettyClassFathers(instance)
        printf((colorFormat.b%'Classes: \n'+'└── '+s+''))
    
    printf((colorFormat.b%'Attrs: '))
    if isinstance(pattern, int):
        deep = pattern
        pattern = None
    dirs = dir(instance)
    if pattern is not None:
        import re
        dirs = [name for name in dirs if re.search(pattern,name)]
        printf('Filter by pattern: "%s"'%(colorFormat.r%pattern))
    def getAttr(attr):
        try:
            try :
                return getattr(instance, attr)
            except :
                pass
            try:
    #            if '__getattribute__' in dirs or 1:
                return instance.__getattribute__(attr)
            except (TypeError): # may be type
                try :
                    return instance.__getattribute__(instance, attr)
                except :
                    pass
                
            except (AttributeError, KeyError):
                pass
            try:
                if '__getattr__' in dirs:
                    return instance.__getattr__(attr)
            except (AttributeError, TypeError):
                return colorFormat.p % '【"getattr"/"getattribute" are not work】'
            return  colorFormat.p % '【No "getattr" or "getattribute"】'
        except Exception as e:
            return colorFormat.p % '【"getAttr" fail, %s(%s)】'%(typestr(e),e)
    l = list(map(getAttr,dirs))
    l = list(map(filterMethodName,dirs,l))
    dic = dict(list(zip(dirs,l)))    
    __dira(dic,instance=instance, maxDocLen=maxDocLen, deep=deep,__key=typeNameOf(type(instance)), printf=printf, )

dira = FunAddMagicMethod(dira)
treea = dira

def what(anything, full=False):
    '''
    tell you what's this by
    pretty print `Document`, `Classes`, `Inner Struct`, `Attributes` of anything.
    
    a magic tool to learn new Python Package

    Parameters
    ----------
    anything : anything in Python
        decompose all of anything
    full : bool, default False
        in default, print lines is less than 10
        set full to True, print all lines
    '''
    tostr = discrib(anything, maxline=not(full) and 10)
    doc = getDoc(anything) or clf.p%"【Not found document】"
    doc = discrib(doc, maxline=not(full) and 15)
    classes = prettyClassFathers(anything)
    doStr = print
    if pyi.jn:
        strs = []
        doStr = strs.append
        
    doStr('-'*10+clf.b%'end of what(' + clf.p%('"%s"'%shortStr(tostr, 30)) + clf.b%')'+'-'*10)
    diraPrintf = PrintStrCollect()
    dira(anything, deep=2, printClassFathers=False, printf=diraPrintf)
    doStr(diraPrintf)
    doStr("")
    doStr((colorFormat.b%'Document: \n'+'└── '+tabstr(doc, 5)+'\n'))    

    innerStruct = isinstance(anything, (list,tuple,dict)) or (typestr(anything) in StructLogFuns)
    if innerStruct:
        treePrintf = PrintStrCollect()
        tree(anything, maxprint=not(full) and 12, printf=treePrintf)
        doStr((colorFormat.b%'Inner Struct:\n')+tounicode(treePrintf))
        doStr("")
        
    doStr((colorFormat.b%'Classes: \n'+'└── '+classes+'\n'))    
    doStr((colorFormat.b%'To Str: \n'+'└── "'+tabstr(tostr, 5)+'"\n'))
    if pyi.jn:
        [print(s) for s in strs[::-1]]
what = FunAddMagicMethod(what)
wtf = what

def __readShape(n):
    return imread(n).shape
def getShapes(imgGlob, returnn=False):
    '''
    以多线程获得匹配imgGlob的所有图片的shapes
    并print出shapes的最大最小值
    
    Parameters
    ----------
    imgGlob : str
        图片的匹配路径
        如 "/img_data/*.png"
    returnn : bool ,default False
        默认不返回shapes
        为True，则返回shapes
    '''
    names = glob.glob(imgGlob)
    shapes = mapmp(__readShape,names)
    lens = list(map(len,shapes))
    dims = np.unique(lens)
    arr = shapes
    print('Dims:')
    for dim in dims:
        print('\t %s dim: %d'%(dim,lens.count(dim)))
    if len(dims)!=1:
        maxx = max(dims)
        arr=[s+(-1,)*(maxx-len(s)) for s in shapes]
    arr=np.array(arr)
    maxx, minn = [],[]
    for dim in range(max(dims)):
        a = arr[:,dim]
        maxx.append(a[a!=-1].max())
        minn.append(a[a!=-1].min())
    
    print('Shape:\n \t max shape: %s\n \t min shape: %s'%(maxx,minn))
    if returnn:
        return shapes

__color10 = [
 (1.0, 1.0, 1.0),
 (0.8666666666666667, 1.0, 1.0),
 (0.8, 1.0, 1.0),
 (0.6666666666666667, 1.0, 1.0),
 (0.5333333333333333, 1.0, 1.0),
 (0.4666666666666667, 1.0, 1.0),
 (0.33333333333333337, 1.0, 1.0),
 (0.19999999999999996, 1.0, 1.0),
 (0.1333333333333333, 1.0, 1.0),
 (0.06666666666666665, 1.0, 1.0)]

def getHsvColors(hn,sn,vn):
    '''
    在hsv空间产生尽可能不同的颜色集
    hn,sn,vn 分别表示 h，s，v对应每个通道的可选值的数目
    return `hn*sn*vn`个 以float三元组为HSV颜色形式的list
    '''
    def toNcolor(n,incloud0=False):
        if incloud0:
            n -= 1
        l = [1.-i*1./n for i in range(n)]
        if incloud0:
            return l+[0.]
        return l
    
    hs,ss,vs = list(map(toNcolor,[hn,sn,vn]))
    cs = []
    for s in ss:
        for v in vs:
            for h in hs:
                c = (h,s,v)
                cs += [c]
    return cs

def getDefaultColorList(colorNum=None, includeBackGround=None,uint8=False):
    '''
    产生尽可能不同的颜色集，用于多label上色
    
    Parameters
    ----------
    colorNum : int, default None
        颜色数目,default 21
    includeBackGround : bool or int, default None
        None: 没有背景色 即不包含黑色
        1 : 第一类为背景即黑色 
        -1: 最后一类为背景即黑色
    uint8 : bool, default False
        是否返还 np.uint8 格式的颜色
        
    Return
    ----------
    以float三元组为RGB颜色形式的list
    '''
    if colorNum is None:
        colorNum=21
        includeBackGround=1
    if includeBackGround is not None:
        colorNum -= 1
    
    if colorNum <= 12:
        colors = getHsvColors(2, 3, 3)
    elif colorNum <= 20:
        colors = __color10+[(c[0],1.0,.5) for c in __color10]
    elif colorNum <= 30:
        colors = __color10+[(c[0],1.0,.66666) for c in __color10]+[(c[0],1.0,.333333) for c in __color10]
        
    elif colorNum <= 60:
        colors = getHsvColors(colorNum//3+1 if colorNum%3 else colorNum//3,1,3)
    else :
        colors = getHsvColors(colorNum//6+1,2,3)
    
    if includeBackGround == -1:
        colors = colors[:colorNum] + [(0.,0.,0.)]
    elif includeBackGround is not None:
        colors = [(0.,0.,0.)] + colors[:colorNum]
    else:
        colors = colors[:colorNum] 
    if uint8 :
        return list((np.array(colors)*255).astype(np.uint8))
    
    from colorsys import hsv_to_rgb
    toRgb = lambda hsv:hsv_to_rgb(*hsv)
    return list(map(toRgb,colors))

def labelToColor(label, colors=None, includeBackGround=None):
    '''
    将颜色集colors映射到label上 返回彩色的label图

    Parameters
    ----------
    label : W*H of int
        W*H的labelMap
    colors : list of RGB color, default None
        对应k类的RGB颜色集，若为None 则会自动生成尽可能不同的颜色集
    includeBackGround : bool or int, default None
        若colors为None 情况下才生效:
            None: 生成的颜色集没有背景色 即不包含黑色
            1   : 第一类为背景即黑色 
            -1  : 最后一类为背景即黑色
    '''
    if colors is None:
        colors = getDefaultColorList(label.max()+1, includeBackGround=includeBackGround)
    colors = np.array(colors)
    if 1.1>(colors).max()>0:
        colors = uint8(colors)
    rgb = np.zeros(label.shape+(3,), np.uint8)
    for c in np.unique(label):
        rgb[label==c] = colors[int(c)]
    return rgb

def standImg(img):
    '''
    任何输入img都转换为 shape is (m, n, 3) dtype == Float
    '''
    from ..ylnp import isNumpyType
    if img.ndim == 2:
        img = greyToRgb(img)
    if img.dtype == np.uint8:
        img = img/255.
    if isNumpyType(img, bool):
        img = img*1.0
    if isNumpyType(img, float):
        return img
    
def getMeanStd(imgPaths):
    '''
    map reduce的方式 获取 所有uint8图片的 mean 和 std
    '''
    paths = imgPaths
    reduceDim = lambda img,axis=0:reduce(lambda x,y:np.append(x,y,axis),img)
    logg = LogLoopTime(paths)
    n = 0
    squre = np.float128([0,0,0])
    summ = np.float128([0,0,0])
    for p in paths:
        pixs = reduceDim(imread(p))/255.
        n += len(pixs)
        squre += (pixs**2).sum(0)
        summ += pixs.sum(0)
        logg(p)
    mean = summ/float(n)
    var = (squre/float(n) - mean**2)
    std = var**.5
    return mean, std

    
if __name__ == '__main__':

    pass



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
A module provide system info and Python Info for boxx

@author: yanglei
"""

import sys, os
from os import environ

def jupyterNotebookOrQtConsole():
    env = 'Unknow'
    cmd = 'ps -ef'
    try:
        with os.popen(cmd) as stream:
            if not py2:
                stream = stream._stream
            s = stream.read()
        pid = os.getpid()
        ls = list(filter(lambda l:'jupyter' in l and str(pid) in l.split(' '), s.split('\n')))
        if len(ls) == 1:
            l = ls[0]
            import re
            pa = re.compile(r'kernel-([-a-z0-9]*)\.json')
            rs = pa.findall(l)
            if len(rs):
                r = rs[0]
                if len(r)<12:
                    env = 'qtipython'
                else :
                    env = 'jn'
        return env
    except:
        return env
    print(r, env)

pyv = sys.version_info.major
py3 = (pyv == 3)
py2 = (pyv == 2)


linuxYl = sys.platform.startswith('linux')
winYl = sys.platform.startswith('win')
osxYl = sys.platform.startswith('darwin')

import multiprocessing as __module
cpun = __module.cpu_count()

cloud = cpun > 8

if linuxYl or osxYl:
    cuda = not os.system('nvidia-smi> /dev/null 2>&1')
elif winYl:
    import subprocess
    try:
        cuda = not subprocess.call('nvidia-smi', creationflags=0x00000008)
    except FileNotFoundError:
        cuda = False
usecuda = 'auto' # auto: auto, False: not use
    
if linuxYl or osxYl:
    homeYl = os.getenv('HOME') + '/'
    tmpYl = '/tmp/'
elif winYl:
    homeYl = os.path.expanduser("~")
    tmpYl = os.getenv('TMP') + '\\'

if "MXX_TMP_DIR" in os.environ:
    tmpYl = os.environ["MXX_TMP_DIR"]

class __TmpboxxWithCall(str):
    '''
    the tmp dir for boxx 
    use tmpboxx() to get tmpdir 
    use tmpboxx({dirname}) to get tmpdir/{dirname}
    if not exist then will auto mkdir of boxxTmp in `/tmp`
    '''
    def __call__(self, dirName=None):
        dirr = os.path.join(self, dirName) if dirName else self
        
        if not os.path.isdir(dirr):
            os.makedirs(dirr, exist_ok=True)
        return dirr
tmpboxx = __TmpboxxWithCall(os.path.join(tmpYl,'boxxTmp/'))

class PythonInfo():
    '''
    python info
    
    plt : Bool
        mean plt avaliable
    env :
        belong [cmd, cmdipython, qtipython, spyder, jn]
    '''
    pid = os.getpid()
    gui = 'ipykernel' in sys.modules
    cmdipython = 'IPython' in sys.modules and not gui
    ipython = cmdipython or gui
    spyder = 'spyder' in sys.modules
    if gui:
        env = 'spyder' if spyder else jupyterNotebookOrQtConsole()
    else:
        env = 'cmdipython' if ipython else 'cmd'
    
    cmd = not ipython
    qtipython = env == 'qtipython'
    jn = env == 'jn'
    
    interactive = bool(getattr(sys, 'ps1', sys.flags.interactive))
    
    plt = True
    if not gui and linuxYl and 'DISPLAY' not in os.environ :
        plt =  False
    reloadplt = False
    def __str__(self):
        from boxx import strMethodForDiraAttrs
        return strMethodForDiraAttrs(self)
    __repr__ = __str__
pyi = PythonInfo()

class SystemInfo():
    '''
    sys info
    '''
    pyv = pyv
    cpun = cpun
    cuda = cuda
    tmp = tmpYl
    
    linux = linuxYl
    win = winYl
    osx = osxYl
    
    os = sys.platform
    display = True
    if linuxYl:
        display = 'DISPLAY' in environ and environ['DISPLAY']
        from os.path import expanduser
        home = expanduser('~')
    gui = pyi.gui or display
    
    @staticmethod
    def ip_by_target_ip(target_ip="8.8.8.8"):
        import socket
    
        return [
            (s.connect((target_ip, 80)), s.getsockname()[0], s.close())
            for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]
        ][0][1]

    @property
    def ip(self):
        return self.ip_by_target_ip()

    @property
    def user(self):
        try:
            import getpass
            return getpass.getuser()
        except KeyError:
            return "unknow"
            
    @property
    def host(self):
        import platform
        return platform.node()
    def __str__(self):
        attrs = {k:str(getattr(self, k)) for k in dir(self) 
                 if not k.startswith('__') and not callable(getattr(self, k))}
        return "\n".join(map(lambda items:"\t%s: %s"%items, attrs.items()))
    __repr__ = __str__
sysi = SystemInfo()
